##########################################################################
# Copyright (C) 2016-2017  Shuaib Osman (sosman@investec.co.za)
# This file is part of RiskFlow.
#
# RiskFlow is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# RiskFlow is distributed in the hope that it will be usefugather_weighted_curvel,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RiskFlow.  If not, see <http://www.gnu.org/licenses/>.
##########################################################################

# utility functions and constants
import utils
# specific modules
import numpy as np
import tensorflow as tf
from collections import OrderedDict

# useful constants
BARRIER_UP = -1.0
BARRIER_DOWN = 1.0
BARRIER_IN = -1.0
BARRIER_OUT = 1.0
OPTION_PUT = -1.0
OPTION_CALL = 1.0


def cash_settle(shared, currency, time_index, value):
    if shared.t_Cashflows is not None:
        shared.t_Cashflows[currency][time_index] += value


class SensitivitiesEstimator(object):
    """ Implements the AAD sensitivities (both first and second derivatives)
       
    Attributes:
        
        value: function output (tensor)
        params: List of model parameters (list of tensor(s))

    """
    def __init__(self, value, params):
        """
        Args:
            cost: Cost function output (tensor)
            params: List of model parameters (list of tensor(s))
        """
        self.cost = value
        self.grad = OrderedDict([(
        var, tf.convert_to_tensor(grad)) for grad, var in zip(
            tf.gradients(value, params, aggregation_method=tf.AggregationMethod.EXPERIMENTAL_ACCUMULATE_N),
            params) if grad is not None])
        self.params = list(self.grad.keys())
        self.P = self.flatten(self.params).get_shape().as_list()[0]

    def flatten(self, params):
        """
        Flattens the list of tensor(s) into a 1D tensor
        
        Args:
            params: List of model parameters (List of tensor(s))
        
        Returns:
            A flattened 1D tensor
        """
        return tf.concat([tf.reshape(_params, [-1]) for _params in params], axis=0)

    def get_Hv_op(self, v):
        """ 
        Implements a Hessian vector product estimator Hv op defined as the 
        matrix multiplication of the Hessian matrix H with the vector v.
    
        Args:      
            v: Vector to multiply with Hessian (tensor)
        
        Returns:
            Hv_op: Hessian vector product op (tensor)
        """
        cost_gradient = self.flatten(self.grad.values())
        vprod = tf.multiply(cost_gradient, tf.stop_gradient(v))
        Hv_op = self.flatten(tf.gradients(vprod, self.params))
        return Hv_op

    def get_H_op(self, prec):
        """ 
        Implements a full Hessian estimator op by forming p Hessian vector 
        products using HessianEstimator.get_Hv_op(v) for all v's in R^P
        
        Args:
            None
        
        Returns:
            H_op: Hessian matrix op (tensor)
        """
        H_op = tf.map_fn(self.get_Hv_op, tf.eye(self.P, self.P, dtype=prec), dtype=prec)
        return H_op


def greeks(shared, deal_data, mtm):    
    greeks_calc = SensitivitiesEstimator(mtm, shared.calc_greeks)
    deal_data.Calc_res['Greeks_First'] = greeks_calc.grad
    # use this only when all the vols and curves are sparsely represented (check greeks_calc.P)
    deal_data.Calc_res['Greeks_Second'] = greeks_calc.get_H_op(shared.precision)


def interpolate(mtm, shared, time_grid, deal_data):
    if deal_data.Time_dep.interp.size > deal_data.Time_dep.deal_time_grid.size:
        # interpolate it
        mtm = utils.gather_interp_matrix(mtm, deal_data.Time_dep, shared)

    # check if we want to store the mtm value for this instrument
    if deal_data.Calc_res is not None:
        deal_data.Calc_res['Value'] = mtm
        if shared.calc_greeks is not None:
            greeks(shared, deal_data, mtm)

    if time_grid.mtm_time_grid.size > 1:
        # pad it with zeros and return
        return tf.pad(mtm, [[0, time_grid.mtm_time_grid.size - deal_data.Time_dep.interp.size], [0, 0]])
    else:
        return mtm


def getbarrierpayoff(direction, eta, phi, strike, barrier):
    '''
    Function to generate the barrier payoff function using these formulae:
    (import sympy with the necessary symbols to see how to derive the code below)
    
    A = phi * ( spot * sympy.exp ( (b-r)*expiry ) * normcdf ( phi * x1 ) -
            strike * sympy.exp( -r*expiry ) * normcdf ( phi * ( x1 - vol ) ) )
    B = phi * ( spot * sympy.exp ( (b-r)*expiry ) * normcdf ( phi * x2 ) -
            strike * sympy.exp( -r*expiry ) * normcdf ( phi * ( x2 - vol ) ) )
    C = phi * ( spot * sympy.exp ( (b-r)*expiry + log_bar*2*(mu+1) ) * normcdf (eta*y1) -
            strike * sympy.exp ( -r*expiry + log_bar*2*mu ) * normcdf ( eta * ( y1 - vol ) ) )
    D = phi * ( spot * sympy.exp ( (b-r)*expiry + log_bar*2*(mu+1) ) * normcdf (eta*y2) -
            strike * sympy.exp ( -r*expiry + log_bar*2*mu ) * normcdf ( eta * ( y2 - vol ) ) )
            
    E = cash_rebate * sympy.exp ( -r*expiry ) * ( normcdf ( eta * ( x2 - vol ) ) -
        sympy.exp(log_bar*2*mu) * normcdf ( eta * ( y2 - vol ) ) )
    F = cash_rebate * ( sympy.exp ( log_bar*(mu+lam) ) * normcdf (eta*z) +
        sympy.exp (log_bar*(mu-lam)) * normcdf ( eta * ( z - 2*lam*vol) ) )

    This is for single Barrier options and based on Merton, Reiner and Rubinstein.
    '''

    def barrier_option(sigma, expiry, cash_rebate, b, r, spot):

        sigma2 = sigma * sigma
        vol = sigma * np.sqrt(expiry)
        mu = (b - 0.5 * sigma2) / sigma2
        log_bar = tf.log(barrier / spot)
        x1 = tf.log(spot / strike) / vol + (1.0 + mu) * vol
        x2 = tf.log(spot / barrier) / vol + (1.0 + mu) * vol

        y1 = tf.log((barrier * barrier) / (spot * strike)) / vol + (1.0 + mu) * vol
        y2 = log_bar / vol + (1.0 + mu) * vol
        lam = tf.sqrt(mu * mu + 2.0 * r / sigma2)
        z = log_bar / vol + lam * vol
        eta_scale = 0.7071067811865476 * eta
        phi_scale = 0.7071067811865476 * phi
        expiry_r = expiry * r

        if direction == BARRIER_IN:
            if ((phi == OPTION_CALL and eta == BARRIER_UP and strike > barrier) or
                    (phi == OPTION_PUT and eta == BARRIER_DOWN and strike <= barrier)):
                # A+E
                return (cash_rebate * (
                        (0.5 * tf.erfc(eta_scale * (-vol + y2)) - 1.0) * tf.exp(2 * log_bar * mu) +
                        0.5 * tf.erfc(eta_scale * (vol - x2))) - phi * (
                                spot * (0.5 * tf.erfc(phi_scale * x1) - 1.0) * tf.exp(b * expiry) +
                                0.5 * strike * tf.erfc(phi_scale * (vol - x1)))) * tf.exp(-expiry_r)
            elif ((phi == OPTION_CALL and eta == BARRIER_UP and strike <= barrier) or
                  (phi == OPTION_PUT and eta == BARRIER_DOWN and strike > barrier)):
                # B-C+D+E
                return (cash_rebate * (
                        (0.5 * tf.erfc(eta_scale * (-vol + y2)) - 1.0) * tf.exp(2 * log_bar * mu) +
                        0.5 * tf.erfc(eta_scale * (vol - x2))) - phi * (
                                spot * (0.5 * tf.erfc(phi_scale * x2) - 1.0) * tf.exp(b * expiry) +
                                0.5 * strike * tf.erfc(phi_scale * (vol - x2))) + phi * (
                                spot * (0.5 * tf.erfc(eta_scale * y1) - 1.0) * tf.exp(
                            expiry * (b - r) + 2 * log_bar * (mu + 1)) - spot * (
                                        0.5 * tf.erfc(eta_scale * y2) - 1.0) *
                                tf.exp(expiry * (b - r) + 2 * log_bar * (mu + 1)) + 0.5 * strike * tf.exp(
                            -expiry_r + 2 * log_bar * mu) * tf.erfc(eta_scale * (vol - y1)) -
                                0.5 * strike * tf.exp(-expiry_r + 2 * log_bar * mu) * tf.erfc(
                            eta_scale * (vol - y2))) * tf.exp(expiry_r)) * tf.exp(-expiry_r)
            elif ((phi == OPTION_PUT and eta == BARRIER_UP and strike > barrier) or
                  (phi == OPTION_CALL and eta == BARRIER_DOWN and strike <= barrier)):
                # A-B+D+E
                return (cash_rebate * ((0.5 * tf.erfc(eta_scale * (-vol + y2)) - 1.0) * tf.exp(
                    2 * log_bar * mu) + 0.5 * tf.erfc(eta_scale * (vol - x2))) - phi * (
                                spot * (0.5 * tf.erfc(eta_scale * y2) - 1.0) * tf.exp(expiry * (
                                b - r) + 2 * log_bar * (mu + 1)) + 0.5 * strike * tf.exp(
                            -expiry_r + 2 * log_bar * mu) * tf.erfc(
                            eta_scale * (vol - y2))) * tf.exp(expiry_r) - phi * (spot * (0.5 * tf.erfc(
                    phi_scale * x1) - 1.0) * tf.exp(b * expiry) + 0.5 * strike * tf.erfc(
                    phi_scale * (vol - x1))) + phi * (spot * (0.5 * tf.erfc(
                    phi_scale * x2) - 1.0) * tf.exp(b * expiry) + 0.5 * strike * tf.erfc(
                    phi_scale * (vol - x2)))) * tf.exp(-expiry_r)
            elif ((phi == OPTION_PUT and eta == BARRIER_UP and strike <= barrier) or
                  (phi == OPTION_CALL and eta == BARRIER_DOWN and strike > barrier)):
                # C+ E
                return (cash_rebate * ((0.5 * tf.erfc(eta_scale * (-vol + y2)) - 1.0) * tf.exp(
                    2 * log_bar * mu) + 0.5 * tf.erfc(eta_scale * (vol - x2))) - phi * (spot * (
                        0.5 * tf.erfc(eta_scale * y1) - 1.0) * tf.exp(expiry * (b - r) + 2 * log_bar * (
                        mu + 1)) + 0.5 * strike * tf.exp(-expiry_r + 2 * log_bar * mu) * tf.erfc(eta_scale * (
                        vol - y1))) * tf.exp(expiry_r)) * tf.exp(-expiry_r)
        else:
            if ((phi == OPTION_CALL and eta == BARRIER_UP and strike > barrier) or
                    (phi == OPTION_PUT and eta == BARRIER_DOWN and strike <= barrier)):
                # F
                return -cash_rebate * ((0.5 * tf.erfc(eta_scale * z) - 1.0) * tf.exp(2 * lam * log_bar) -
                                       0.5 * tf.erfc(eta_scale * (2 * lam * vol - z))) * tf.exp(-log_bar * (lam - mu))
            elif ((phi == OPTION_CALL and eta == BARRIER_UP and strike <= barrier) or
                  (phi == OPTION_PUT and eta == BARRIER_DOWN and strike > barrier)):
                # A - B + C - D + F
                return (-cash_rebate * ((0.5 * tf.erfc(eta_scale * z) - 1.0) * tf.exp(2 * lam * log_bar)
                                        - 0.5 * tf.erfc(eta_scale * (2 * lam * vol - z))) * tf.exp(expiry_r)
                        + phi * (-spot * (0.5 * tf.erfc(eta_scale * y1) - 1.0) * tf.exp(
                            expiry * (b - r) + 2 * log_bar * (mu + 1)) + spot * (0.5 * tf.erfc(eta_scale * y2) - 1.0) *
                                 tf.exp(expiry * (b - r) + 2 * log_bar * (mu + 1)) - 0.5 * strike * tf.exp(
                                    -expiry_r + 2 * log_bar * mu) *
                                 tf.erfc(eta_scale * (vol - y1)) + 0.5 * strike * tf.exp(-expiry_r + 2 * log_bar * mu) *
                                 tf.erfc(eta_scale * (vol - y2))) * tf.exp(expiry_r + log_bar * (lam - mu)) +
                        phi * (-spot * (0.5 * tf.erfc(phi_scale * x1) - 1.0) * tf.exp(b * expiry) + spot * (
                                0.5 * tf.erfc(phi_scale * x2) - 1.0) * tf.exp(b * expiry) - 0.5 * strike *
                               tf.erfc(phi_scale * (vol - x1)) + 0.5 * strike * tf.erfc(phi_scale * (vol - x2)))
                        * tf.exp(log_bar * (lam - mu))) * tf.exp(-expiry_r - log_bar * (lam - mu))
            elif ((phi == OPTION_PUT and eta == BARRIER_UP and strike > barrier) or
                  (phi == OPTION_CALL and eta == BARRIER_DOWN and strike <= barrier)):
                # B - D + F
                return (-cash_rebate * ((0.5 * tf.erfc(eta_scale * z) - 1.0) * tf.exp(2 * lam * log_bar)
                                        - 0.5 * tf.erfc(eta_scale * (2 * lam * vol - z))) * tf.exp(expiry_r) +
                        phi * (spot * (0.5 * tf.erfc(eta_scale * y2) - 1.0) * tf.exp(
                            expiry * (b - r) + 2 * log_bar * (mu + 1))
                               + 0.5 * strike * tf.exp(-expiry_r + 2 * log_bar * mu) * tf.erfc(
                                    eta_scale * (vol - y2))) * tf.exp(
                            expiry_r + log_bar * (lam - mu)) - phi * (
                                spot * (0.5 * tf.erfc(phi_scale * x2) - 1.0) * tf.exp(
                            b * expiry) + 0.5 * strike * tf.erfc(phi_scale * (vol - x2))) * tf.exp(
                            log_bar * (lam - mu))) * tf.exp(
                    -expiry_r - log_bar * (lam - mu))
            elif ((phi == OPTION_PUT and eta == BARRIER_UP and strike <= barrier) or
                  (phi == OPTION_CALL and eta == BARRIER_DOWN and strike > barrier)):
                # A-C+F
                return (-cash_rebate * ((0.5 * tf.erfc(eta_scale * z) - 1.0) * tf.exp(2 * lam * log_bar)
                                        - 0.5 * tf.erfc(eta_scale * (2 * lam * vol - z))) * tf.exp(expiry_r) +
                        phi * (spot * (0.5 * tf.erfc(eta_scale * y1) - 1.0) * tf.exp(
                            expiry * (b - r) + 2 * log_bar * (mu + 1))
                               + 0.5 * strike * tf.exp(-expiry_r + 2 * log_bar * mu) * tf.erfc(
                                    eta_scale * (vol - y1))) * tf.exp(
                            expiry_r + log_bar * (lam - mu)) - phi * (
                                spot * (0.5 * tf.erfc(phi_scale * x1) - 1.0) * tf.exp(
                            b * expiry) + 0.5 * strike * tf.erfc(phi_scale * (vol - x1))) * tf.exp(
                            log_bar * (lam - mu))) * tf.exp(
                    -expiry_r - log_bar * (lam - mu))

    return barrier_option


def pvbarrieroption(shared, time_grid, deal_data, nominal,
                    spot, b, tau, payoff_currency, invert_moneyness=False):
    deal_time = time_grid.time_grid[deal_data.Time_dep.deal_time_grid]

    factor_dep = deal_data.Factor_dep
    daycount_fn = factor_dep['Discount'][0][utils.FACTOR_INDEX_Daycount]

    # work out what we're pricing
    phi = OPTION_CALL if deal_data.Instrument.field['Option_Type'] == 'Call' else OPTION_PUT
    eta = BARRIER_DOWN if 'Down' in deal_data.Instrument.field['Barrier_Type'] else BARRIER_UP
    direction = BARRIER_OUT if 'Out' in deal_data.Instrument.field['Barrier_Type'] else BARRIER_IN
    buy_or_sell = 1.0 if deal_data.Instrument.field['Buy_Sell'] == 'Buy' else -1.0
    barrier = deal_data.Instrument.field['Barrier_Price']
    strike = deal_data.Instrument.field['Strike_Price']
    cash_rebate = deal_data.Instrument.field['Cash_Rebate']

    # get the zero curve
    discounts = utils.calc_time_grid_curve_rate(factor_dep['Discount'], deal_time[:-1], shared)

    expiry = daycount_fn(tau).astype(shared.precision)
    need_spot_at_expiry = deal_time.shape[0] - expiry.size
    spot_prior, spot_at = tf.split(spot, [expiry.size, need_spot_at_expiry])
    moneyness = strike / spot_prior if invert_moneyness else spot_prior / strike
    sigma = utils.calc_time_grid_vol_rate(factor_dep['Volatility'], moneyness, expiry, shared)

    if factor_dep['Barrier_Monitoring']:
        adj_barrier = barrier * tf.exp((2.0 * tf.cast(
            barrier > spot, shared.precision) - 1.0) * sigma * factor_dep['Barrier_Monitoring'])
    else:
        adj_barrier = barrier

    # get the payoff function - should be adj_barrier not barrier - TODO
    barrierOption = getbarrierpayoff(direction, eta, phi, strike, barrier)

    r = tf.squeeze(discounts.gather_weighted_curve(
        shared, tau.reshape(-1, 1), multiply_by_time=False), axis=1)

    expiry_years = expiry.reshape(-1, 1)
    barrier_payoff = buy_or_sell * nominal * barrierOption(sigma, expiry_years, cash_rebate / nominal,
                                                           b, r, spot_prior)

    if need_spot_at_expiry:
        # work out barrier
        if eta == BARRIER_UP:
            touched = tf.cast(tf.logical_and(
                spot[:-1] < barrier, spot[1:] > barrier), shared.precision)
        else:
            touched = tf.cast(tf.logical_and(
                spot[:-1] > barrier, spot[1:] < barrier), shared.precision)

        # barrier payoff
        barrier_touched = tf.pad(tf.cast(tf.cumsum(touched, axis=0) > 0, shared.precision), [[1, 0], [0, 0]])
        first_touch = barrier_touched[1:] - barrier_touched[:-1]
        # final payoff
        payoff_at = buy_or_sell * tf.nn.relu(phi * (spot_at - strike))

        if direction == BARRIER_IN:
            forward = spot_prior * tf.exp(b * expiry_years)
            payoff_prior = utils.black_european_option(
                forward, strike, sigma, expiry, buy_or_sell, phi, shared) * tf.exp(-r * expiry_years)
            european_part = barrier_touched * (nominal * tf.concat([payoff_prior, payoff_at], axis=0))
            barrier_part = (1.0 - barrier_touched) * tf.pad(
                barrier_payoff, [[0, 1], [0, 0]], constant_values=buy_or_sell * cash_rebate)
            combined = european_part + barrier_part
            # settle cashflows (can only happen at the end)
            cash_settle(shared, payoff_currency, deal_data.Time_dep.deal_time_grid[-1], combined[-1])
        else:
            # barrier out
            barrier_part = (1.0 - barrier_touched) * tf.concat([barrier_payoff, nominal * payoff_at], axis=0)
            rebate_part = buy_or_sell * cash_rebate * first_touch
            combined = tf.pad(buy_or_sell * cash_rebate * first_touch, [[1, 0], [0, 0]]) + barrier_part
            # settle cashflows (The one at expiry)
            cash_settle(shared, payoff_currency, deal_data.Time_dep.deal_time_grid[-1], barrier_part[-1])
            # settle cashflows (The potential rebate knockout)
            if cash_rebate:
                for cash_index, cash in zip(deal_data.Time_dep.deal_time_grid[1:], tf.unstack(rebate_part)):
                    cash_settle(shared, payoff_currency, cash_index, cash)
    else:
        combined = barrier_payoff

    return combined


def pvonetouchoption(shared, time_grid, deal_data, nominal,
                     spot, b, tau, payoff_currency, invert_moneyness=False):
    factor_dep = deal_data.Factor_dep
    deal_time = time_grid.time_grid[deal_data.Time_dep.deal_time_grid]
    daycount_fn = factor_dep['Discount'][0][utils.FACTOR_INDEX_Daycount]

    # work out what we're pricing

    eta = BARRIER_DOWN if 'Down' in deal_data.Instrument.field['Barrier_Type'] else BARRIER_UP
    buy_or_sell = 1.0 if deal_data.Instrument.field['Buy_Sell'] == 'Buy' else -1.0
    barrier = deal_data.Instrument.field['Barrier_Price']

    # get the zero curve
    discounts = utils.calc_time_grid_curve_rate(factor_dep['Discount'], deal_time[:-1], shared)

    expiry = daycount_fn(tau).astype(shared.precision)
    need_spot_at_expiry = deal_time.shape[0] - expiry.size
    spot_prior, spot_at = tf.split(spot, [expiry.size, need_spot_at_expiry])
    moneyness = barrier / spot_prior if invert_moneyness else spot_prior / barrier
    sigma = utils.calc_time_grid_vol_rate(factor_dep['Volatility'], moneyness, expiry, shared)

    if factor_dep['Barrier_Monitoring']:
        adj_barrier = barrier * tf.exp((2.0 * tf.cast(
            barrier > spot_prior, shared.precision) - 1.0) * sigma * factor_dep['Barrier_Monitoring'])
    else:
        adj_barrier = barrier

    r = tf.squeeze(discounts.gather_weighted_curve(shared, tau.reshape(-1, 1), multiply_by_time=False), axis=1)

    # one touch payoff
    expiry_years = expiry.reshape(-1, 1)
    root_tau = np.sqrt(expiry_years)
    mu = b / sigma - 0.5 * sigma
    log_vol = tf.log(barrier / spot_prior) / sigma
    barrovert = log_vol / root_tau
    eta_scale = 0.7071067811865476 * eta

    if deal_data.Instrument.field['Payment_Timing'] == 'Expiry':
        muroot = mu * root_tau
        d1 = muroot - barrovert
        d2 = -muroot - barrovert
        payoff = tf.exp(-r * expiry_years) * 0.5 * (
                tf.erfc(eta_scale * d1) + tf.exp(2.0 * mu * log_vol) * tf.erfc(eta_scale * d2))
    elif deal_data.Instrument.field['Payment_Timing'] == 'Touch':
        lamb = tf.sqrt(tf.nn.relu(mu * mu + 2.0 * r))
        lambroot = lamb * root_tau
        d1 = lambroot - barrovert
        d2 = -lambroot - barrovert
        payoff = 0.5 * (tf.exp((mu - lamb) * log_vol) * tf.erfc(eta_scale * d1) +
                        tf.exp((mu + lamb) * log_vol) * tf.erfc(eta_scale * d2))

    if need_spot_at_expiry:
        # barrier check
        if eta == BARRIER_UP:
            touched = tf.cast(tf.logical_and(
                spot[:-1] < barrier, spot[1:] > barrier), shared.precision)
        else:
            touched = tf.cast(tf.logical_and(
                spot[:-1] > barrier, spot[1:] < barrier), shared.precision)

        barrier_touched = tf.pad(tf.cast(tf.cumsum(touched, axis=0) > 0, shared.precision), [[1, 0], [0, 0]])
        first_touch = barrier_touched[1:] - barrier_touched[:-1]
        barrier_part = (1.0 - barrier_touched) * tf.pad(payoff, [[0, 1], [0, 0]])

        if deal_data.Instrument.field['Payment_Timing'] == 'Touch':
            touch_part = tf.pad(first_touch, [[1, 0], [0, 0]])
            combined = buy_or_sell * nominal * (touch_part + barrier_part)
            for cash_index, cash in zip(deal_data.Time_dep.deal_time_grid[1:], tf.unstack(first_touch)):
                cash_settle(shared, payoff_currency, cash_index, buy_or_sell * nominal * cash)
        else:
            # Expiry
            rebate_part = barrier_touched * tf.pad(tf.exp(-r * expiry_years), [[0, 1], [0, 0]], constant_values=1.0)
            combined = buy_or_sell * nominal * (rebate_part + barrier_part)
            # settle cashflows (The one at expiry)
            cash_settle(shared, payoff_currency, deal_data.Time_dep.deal_time_grid[-1], combined[-1])
    else:
        combined = buy_or_sell * nominal * payoff

    return combined


def pricer_float_cashflows(all_resets, cashflows, factor_dep, time_slice, shared):
    margin = cashflows[:, utils.CASHFLOW_INDEX_FloatMargin] * cashflows[:, utils.CASHFLOW_INDEX_Year_Frac]
    all_int = cashflows[:, utils.CASHFLOW_INDEX_Year_Frac].reshape(1, -1, 1) * all_resets + margin.reshape(1, -1, 1)

    return all_int, margin


def pricer_cap(all_resets, cashflows, factor_dep, time_slice, shared):
    mn_option = all_resets - cashflows[:, utils.CASHFLOW_INDEX_Strike].reshape(1, -1, 1)
    expiry = factor_dep['Discount'][0][utils.FACTOR_INDEX_Daycount](
        cashflows[:, utils.CASHFLOW_INDEX_Start_Day] - time_slice)

    # note that the Year Frac is averaged - all the cashflows are supposed to have the same year frac
    # (but practically not - should be ok to do this)

    vols = utils.calc_tenor_cap_time_grid_vol_rate(
        factor_dep['VolSurface'], mn_option, expiry,
        cashflows[:, utils.CASHFLOW_INDEX_Year_Frac].mean(), shared)

    payoff = utils.black_european_option(
        all_resets, cashflows[:, utils.CASHFLOW_INDEX_Strike].reshape(1, -1, 1),
        vols, expiry, 1.0, 1.0, shared)

    all_int = cashflows[:, utils.CASHFLOW_INDEX_Year_Frac].reshape(1, -1, 1) * payoff
    margin = np.zeros(cashflows.shape[0], dtype=shared.precision)

    return all_int, margin


def pricer_floor(all_resets, cashflows, factor_dep, time_slice, shared):
    mn_option = all_resets - cashflows[:, utils.CASHFLOW_INDEX_Strike].reshape(1, -1, 1)
    expiry = factor_dep['Discount'][0][utils.FACTOR_INDEX_Daycount](
        cashflows[:, utils.CASHFLOW_INDEX_Start_Day] - time_slice)

    # note that the Year Frac is averaged - all the cashflows are supposed to have the same year frac
    # (but practically not - should be ok to do this)

    vols = utils.calc_tenor_cap_time_grid_vol_rate(
        factor_dep['VolSurface'], mn_option, expiry,
        cashflows[:, utils.CASHFLOW_INDEX_Year_Frac].mean(), shared)

    payoff = utils.black_european_option(
        all_resets, cashflows[:, utils.CASHFLOW_INDEX_Strike].reshape(1, -1, 1),
        vols, expiry, 1.0, -1.0, shared)

    all_int = cashflows[:, utils.CASHFLOW_INDEX_Year_Frac].reshape(1, -1, 1) * payoff
    margin = np.zeros(cashflows.shape[0], dtype=shared.precision)

    return all_int, margin


def pvfloatcashflowlist(shared, time_grid, deal_data, cashflow_pricer, mtm_currency=None, settle_cash=True):
    mtm_list = []
    factor_dep = deal_data.Factor_dep
    deal_time = time_grid.time_grid[deal_data.Time_dep.deal_time_grid]

    # first precalc all past resets
    resets = factor_dep['Cashflows'].Resets
    known_resets = resets.known_resets(shared.simulation_batch)
    sim_resets = resets.schedule[(resets.schedule[:, utils.RESET_INDEX_Scenario] > -1) &
                                 (resets.schedule[:, utils.RESET_INDEX_Reset_Day] <=
                                  deal_time[:, utils.TIME_GRID_MTM].max())]
    if mtm_currency:
        # precalc the FX forwards
        sim_fx_forward = utils.calc_fx_forward(
            mtm_currency, factor_dep['Currency'],
            sim_resets[:, utils.RESET_INDEX_Start_Day], sim_resets[:, :utils.RESET_INDEX_Scenario + 1],
            shared, only_diag=True)

        known_fx = factor_dep['Cashflows'].known_resets(
            shared.simulation_batch, index=utils.CASHFLOW_INDEX_FXResetValue,
            filter_index=utils.CASHFLOW_INDEX_Start_Day)

        # fetch fx rates - note that there is a slight difference between this and the spot fx rate
        old_fx_rates = tf.squeeze(
            tf.concat([tf.stack(known_fx), sim_fx_forward]
                      if known_fx
                      else sim_fx_forward, axis=0), axis=1)

    forwards = utils.calc_time_grid_curve_rate(factor_dep['Forward'], deal_time, shared)
    discounts = utils.calc_time_grid_curve_rate(factor_dep['Discount'], deal_time, shared)
    old_resets = utils.calc_time_grid_curve_rate(factor_dep['Forward'],
                                                 sim_resets[:, :utils.RESET_INDEX_Scenario + 1],
                                                 shared)

    delta_start = (sim_resets[:, utils.RESET_INDEX_Start_Day] -
                   sim_resets[:, utils.RESET_INDEX_Reset_Day]).reshape(-1, 1)
    delta_end = (sim_resets[:, utils.RESET_INDEX_End_Day] -
                 sim_resets[:, utils.RESET_INDEX_Reset_Day]).reshape(-1, 1)

    reset_weights = (sim_resets[:, utils.RESET_INDEX_Weight] /
                     sim_resets[:, utils.RESET_INDEX_Accrual]).reshape(-1, 1, 1)

    reset_values = tf.expm1(old_resets.gather_weighted_curve(
        shared, delta_end, delta_start)) * reset_weights \
        if sim_resets.any() else tf.zeros([0, 1, shared.simulation_batch],
                                          dtype=shared.precision)

    # fetch all fixed resets 
    old_resets = tf.squeeze(
        tf.concat([tf.stack(known_resets), reset_values]
                  if known_resets
                  else reset_values, axis=0), axis=1)

    cash_start_idx = factor_dep['Cashflows'].get_cashflow_start_index(deal_time)
    start_index, start_counts = np.unique(cash_start_idx, return_counts=True)

    for index, (forward_block, discount_block) in enumerate(utils.split_counts(
            [forwards, discounts], start_counts, shared)):

        cashflows = factor_dep['Cashflows'].merged()[start_index[index]:]

        cash_pmts, cash_index, cash_counts = np.unique(cashflows[:, utils.CASHFLOW_INDEX_Pay_Day],
                                                       return_index=True, return_counts=True)

        reset_offset = factor_dep['Cashflows'].offsets[start_index[index]][1]
        pmts_offset = cash_index + (cash_counts - 1)

        time_ofs = 0
        time_block = discount_block.time_grid[:, utils.TIME_GRID_MTM]
        future_pmts = cash_pmts.reshape(1, -1) - time_block.reshape(-1, 1)
        reset_block = resets.schedule[reset_offset:]
        reset_ofs, reset_count = np.unique(resets.split_block_resets(
            reset_offset, time_block), return_counts=True)

        # discount rates
        discount_rates = utils.calc_discount_rate(discount_block,
                                                  future_pmts, shared)

        # do we need to split the forward block further? 
        forward_blocks = forward_block.split_counts(
            reset_count, shared) if len(reset_count) > 1 else [forward_block]

        # empty list for payments
        payments = []

        for offset, size, forward_rates in zip(*[reset_ofs, reset_count, forward_blocks]):
            time_slice = time_block[time_ofs:time_ofs + size].reshape(-1, 1)
            future_starts = reset_block[offset:, utils.RESET_INDEX_Start_Day] - time_slice
            future_ends = reset_block[offset:, utils.RESET_INDEX_End_Day] - time_slice
            future_weights = (reset_block[offset:, utils.RESET_INDEX_Weight]
                              / reset_block[offset:, utils.RESET_INDEX_Accrual]).reshape(1, -1, 1)
            future_resets = tf.expm1(forward_rates.gather_weighted_curve(
                shared, future_ends, future_starts)) * future_weights

            # now deal with past resets
            past_resets = tf.tile(
                tf.expand_dims(old_resets[reset_offset:reset_offset + offset], axis=0)
                , [size, 1, 1])
            all_resets = tf.concat([past_resets, future_resets], axis=1)

            # handle cashflows that don't pay interest (e.g. bullets)
            if cashflows[:, utils.CASHFLOW_INDEX_NumResets].all():
                reset_cashflows = cashflows
            else:
                reset_cash_index = np.where(cashflows[:, utils.CASHFLOW_INDEX_NumResets])[0]
                reset_cashflows = cashflows[reset_cash_index]
                cash_counts *= (cashflows[:, utils.CASHFLOW_INDEX_NumResets] >= 1).astype(np.int32)
                cash_index = reset_cash_index.searchsorted(cash_index)

            if mtm_currency:
                # now deal with fx rates - note that there should only be 1 reset per cashflow
                future_fx_resets = utils.calc_fx_forward(
                    mtm_currency, factor_dep['Currency'],
                    cashflows[offset:, utils.CASHFLOW_INDEX_FXResetDate],
                    discount_block.time_grid[time_ofs:time_ofs + size], shared)

                past_fx_resets = tf.tile(
                    tf.expand_dims(old_fx_rates[reset_offset:reset_offset + offset], axis=0)
                    , [size, 1, 1])
                all_fx_resets = tf.concat([past_fx_resets, future_fx_resets], axis=1)

                # calculate the Nominal in the correct currency
                Pi = all_fx_resets * cashflows[:, utils.CASHFLOW_INDEX_Nominal].reshape(1, -1, 1)
                Pi_1 = tf.pad(Pi[:, 1:, :], [[0, 0], [0, 1], [0, 0]])

            time_ofs += size

            # now we can price the cashflows
            all_int, all_margin = cashflow_pricer(all_resets, reset_cashflows, factor_dep, time_slice, shared)

            # check if there are a different number of resets per cashflow 
            if (cash_counts.min() != cash_counts.max()):
                interest = tf.pad(all_int, [[0, 0], [0, 1], [0, 0]])
                nominal = np.append(reset_cashflows[:, utils.CASHFLOW_INDEX_Nominal], 0.0)
                margin = np.append(all_margin, 0.0)
            else:
                interest = all_int
                margin = all_margin
                nominal = reset_cashflows[:, utils.CASHFLOW_INDEX_Nominal]

            default_offst = np.ones(cash_index.size, dtype=np.int32) * (interest.shape[1].value - 1)
            total = 0.0

            for i in range(cash_counts.max()):
                offst = default_offst.copy()
                offst[cash_counts > i] = i + cash_index[cash_counts > i]
                int_i = tf.gather(interest, offst, axis=1)

                if mtm_currency:
                    total += int_i * Pi + (Pi - Pi_1)
                elif factor_dep['CompoundingMethod'] == 'None':
                    total += nominal[offst].reshape(1, -1, 1) * int_i
                elif factor_dep['CompoundingMethod'] == 'Include_Margin':
                    total += (total + nominal[offst].reshape(1, -1, 1)) * int_i
                elif factor_dep['CompoundingMethod'] == 'Flat':
                    total += (int_i * nominal[offst].reshape(1, -1, 1)) + total * (
                            int_i - margin[offst].reshape(1, -1, 1))
                else:
                    raise Exception('Floating cashflow list method not implemented')

            payments.append(total + cashflows[pmts_offset, utils.CASHFLOW_INDEX_FixedAmt].reshape(1, -1, 1))

        # now finish the payments
        all_payments = tf.concat(payments, axis=0) if len(payments) > 1 else payments[0]

        # settle any cashflows
        if settle_cash:
            cash_settle(shared, factor_dep['SettleCurrency'], np.searchsorted(time_grid.mtm_time_grid, cash_pmts[0]),
                        all_payments[-1][0])
        # add it to the list
        mtm_list.append(tf.reduce_sum(all_payments * discount_rates, axis=1))

    return tf.concat(mtm_list, axis=0, name='mtm')


def pvfixedcashflows(shared, time_grid, deal_data, ignore_fixed_rate=False, settle_cash=True):
    mtm_list = []
    factor_dep = deal_data.Factor_dep
    deal_time = time_grid.time_grid[deal_data.Time_dep.deal_time_grid]

    cash_start_idx = factor_dep['Cashflows'].get_cashflow_start_index(deal_time)
    settlement_amt = factor_dep.get('Settlement_Amount', 0.0)
    discounts = utils.calc_time_grid_curve_rate(factor_dep['Discount'], deal_time, shared)

    start_index, counts = np.unique(cash_start_idx, return_counts=True)

    for index, [discount_block] in enumerate(utils.split_counts([discounts], counts, shared)):
        cashflows = factor_dep['Cashflows'].schedule[start_index[index]:]

        cash_pmts, cash_index, cash_counts = np.unique(cashflows[:, utils.CASHFLOW_INDEX_Pay_Day],
                                                       return_index=True, return_counts=True)
        # payment times            
        time_block = discount_block.time_grid[:, utils.TIME_GRID_MTM]
        future_pmts = cash_pmts.reshape(1, -1) - time_block.reshape(-1, 1)

        # discount rates
        discount_rates = utils.calc_discount_rate(discount_block,
                                                  future_pmts, shared)
        # is this a forward?
        if settlement_amt:
            settlement = settlement_amt * tf.squeeze(utils.calc_discount_rate(
                discount_block,
                (factor_dep['Settlement_Date'] - time_block).reshape(-1, 1), shared), axis=1)
        else:
            settlement = 0.0

        # empty list for payments
        all_int = (1.0 if ignore_fixed_rate else
                   cashflows[:, utils.CASHFLOW_INDEX_FixedRate]) * cashflows[:, utils.CASHFLOW_INDEX_Year_Frac]
        payments = np.zeros_like(cash_index, dtype=shared.precision)

        if cash_counts.min() != cash_counts.max():
            interest = np.append(all_int, 0.0)
            nominal = np.append(cashflows[:, utils.CASHFLOW_INDEX_Nominal], 0.0)
            fixed_amt = np.append(cashflows[:, utils.CASHFLOW_INDEX_FixedAmt], 0.0)
        else:
            interest = all_int
            nominal = cashflows[:, utils.CASHFLOW_INDEX_Nominal]
            fixed_amt = cashflows[:, utils.CASHFLOW_INDEX_FixedAmt]

        default_offst = np.ones(cash_index.size, dtype=np.int32) * (interest.size - 1)

        for i in range(cash_counts.max()):
            offst = default_offst.copy()
            offst[cash_counts > i] = i + cash_index[cash_counts > i]
            int_i = interest[offst]

            if factor_dep.get('Compounding', False):
                payments += (payments + nominal[offst]) * int_i + fixed_amt[offst]
            else:
                payments += int_i * nominal[offst] + fixed_amt[offst]

        # add to the mtm               
        mtm_list.append(tf.reduce_sum(payments.reshape(1, -1, 1) * discount_rates, axis=1) - settlement)

        # settle any cashflows
        if settle_cash:
            if factor_dep.get('Settlement_Date') is not None:
                cash_settle(shared, factor_dep['SettleCurrency'],
                            np.searchsorted(time_grid.mtm_time_grid, factor_dep['Settlement_Date']),
                            mtm_list[-1][-1])
            else:
                cash_settle(shared, factor_dep['SettleCurrency'],
                            np.searchsorted(time_grid.mtm_time_grid, cash_pmts[0]),
                            payments[0])

    return tf.concat(mtm_list, axis=0, name='mtm')


def pvindexcashflows(shared, time_grid, deal_data, settle_cash=True):
    def calc_index(schedule, sim_schedule):
        weight = schedule[:, utils.RESET_INDEX_Weight].reshape(1, -1, 1)
        dates = schedule[np.newaxis, :, utils.RESET_INDEX_Reset_Day] - \
                last_pub_block[:, np.newaxis, utils.RESET_INDEX_Reset_Day]
        index_t = tf.expand_dims(last_index_block, axis=1) / utils.calc_discount_rate(
            forecast_block, dates, shared)

        # split if necessary
        if dates[dates < 0].any():
            future_indices = (dates >= 0).all(axis=1).argmin()
            future_index_t, past_index_t = tf.split(
                index_t, [future_indices, dates.shape[0] - future_indices])

            mixed_indices_t = []
            for mixed_dates, mixed_indices in zip(dates[future_indices:], tf.unstack(past_index_t)):
                future_resets = mixed_dates.size - (mixed_dates[::-1] > 0).argmin()
                past_resets_t, future_resets_t = tf.split(
                    mixed_indices, [future_resets, mixed_dates.size - future_resets])
                mixed_indices_t.append(
                    tf.concat([sim_schedule[:future_resets], future_resets_t], axis=0))

            values = weight * tf.concat([future_index_t, tf.stack(mixed_indices_t)], axis=0)
        else:
            values = weight * index_t

        if resets_per_cf > 1:
            return tf.reduce_sum(
                tf.reshape(values,
                           (last_pub_block.shape[0], -1, resets_per_cf, shared.simulation_batch)), axis=2)
        else:
            return values

    def get_index_val(cash_index_vals, schedule, sim_schedule, resets_per_cf, offset):
        if cash_index_vals[cash_index_vals < 0].any():
            num_known = cash_index_vals[cash_index_vals > 0].size
            known_indices = tf.tile(
                cash_index_vals[cash_index_vals > 0].reshape(1, -1, 1).astype(shared.precision),
                (last_pub_block.shape[0], 1, shared.simulation_batch))
            reset_offset = resets_per_cf * (offset + num_known)
            if num_known:
                return tf.concat(
                    [known_indices,
                     calc_index(schedule[reset_offset:], sim_schedule[reset_offset:])], axis=1)
            else:
                return calc_index(schedule[reset_offset:], sim_schedule[reset_offset:])
        else:
            return cash_index_vals.reshape(1, -1, 1)

    def filter_resets(resets, index):
        known_resets = resets.known_resets(shared.simulation_batch)
        sim_resets = resets.schedule[(resets.schedule[:, utils.RESET_INDEX_Scenario] > -1) &
                                     (resets.schedule[:, utils.RESET_INDEX_Reset_Day] <=
                                      deal_time[:, utils.TIME_GRID_MTM].max())]
        old_resets = utils.calc_time_grid_spot_rate(index, sim_resets[:, :utils.RESET_INDEX_Scenario + 1], shared)
        return tf.concat([tf.concat(known_resets, axis=0), old_resets], axis=0) if known_resets else old_resets

    mtm_list = []
    factor_dep = deal_data.Factor_dep
    deal_time = time_grid.time_grid[deal_data.Time_dep.deal_time_grid]
    cash_start_idx = factor_dep['Cashflows'].get_cashflow_start_index(deal_time)

    resets_per_cf = 2 if (
            factor_dep['IndexMethod'] in [utils.CASHFLOW_METHOD_IndexReferenceInterpolated3M,
                                          utils.CASHFLOW_METHOD_IndexReferenceInterpolated4M]) else 1

    last_published = factor_dep['Cashflows'].Resets.schedule[deal_data.Time_dep.deal_time_grid]
    last_published_index = utils.calc_time_grid_spot_rate(factor_dep['PriceIndex'], deal_time, shared)
    index_forecast = utils.calc_time_grid_curve_rate(factor_dep['ForecastIndex'], deal_time, shared)
    discounts = utils.calc_time_grid_curve_rate(factor_dep['Discount'], deal_time, shared)

    all_base_resets = filter_resets(factor_dep['Base_Resets'], factor_dep['PriceIndex'])
    all_final_resets = filter_resets(factor_dep['Final_Resets'], factor_dep['PriceIndex'])

    start_index, start_counts = np.unique(cash_start_idx, return_counts=True)
    last_pub_blocks = np.split(last_published, start_counts.cumsum())

    for index, (forecast_block, discount_block, last_index_block) in enumerate(
            utils.split_counts([index_forecast, discounts, last_published_index], start_counts, shared)):

        last_pub_block = last_pub_blocks[index]
        cashflows = factor_dep['Cashflows'].schedule[start_index[index]:]

        cash_pmts, cash_index, cash_counts = np.unique(cashflows[:, utils.CASHFLOW_INDEX_Pay_Day],
                                                       return_index=True, return_counts=True)

        # payment times            
        time_block = discount_block.time_grid[:, utils.TIME_GRID_MTM]
        future_pmts = cash_pmts.reshape(1, -1) - time_block.reshape(-1, 1)

        # discount rates
        discount_rates = utils.calc_discount_rate(discount_block,
                                                  future_pmts, shared)

        cash_base_index_vals = factor_dep['Cashflows'].offsets[start_index[index]:, utils.CASHFLOW_OFFSET_BaseReference]
        all_base_index_vals = get_index_val(cash_base_index_vals, factor_dep['Base_Resets'].schedule, all_base_resets,
                                            resets_per_cf, start_index[index])
        cash_final_index_vals = factor_dep['Cashflows'].offsets[start_index[index]:,
                                utils.CASHFLOW_OFFSET_FinalReference]
        all_final_index_vals = get_index_val(cash_final_index_vals, factor_dep['Final_Resets'].schedule,
                                             all_final_resets,
                                             resets_per_cf, start_index[index])

        # empty list for payments
        interest = (
                cashflows[:, utils.CASHFLOW_INDEX_FixedRate] * cashflows[:, utils.CASHFLOW_INDEX_Year_Frac]). \
            reshape(1, -1, 1)
        growth = cashflows[:, utils.CASHFLOW_INDEX_FixedAmt].reshape(1, -1, 1) * (
                all_final_index_vals / all_base_index_vals)
        payment = cashflows[:, utils.CASHFLOW_INDEX_Nominal].reshape(1, -1, 1) * growth * interest
        payments = None

        for i in range(cash_counts.min()):
            offst = cash_index + i
            int_i = tf.cast(tf.gather(payment, offst, axis=1), shared.precision)
            if payments is None:
                payments = int_i
            else:
                payments += int_i

        # add it to the list
        mtm_list.append(tf.reduce_sum(payments * discount_rates, axis=1))

        # settle any cashflows
        if settle_cash:
            if factor_dep['Settlement_Date'] is not None:
                cash_settle(shared, factor_dep['SettleCurrency'],
                            np.searchsorted(time_grid.mtm_time_grid, factor_dep['Settlement_Date']),
                            mtm_list[-1][-1])
            else:
                cash_settle(shared, factor_dep['SettleCurrency'],
                            np.searchsorted(time_grid.mtm_time_grid, cash_pmts[0]),
                            payments[-1][0])

    return tf.concat(mtm_list, axis=0, name='mtm')


def pvenergycashflows(shared, time_grid, deal_data):
    mtm_list = []
    factor_dep = deal_data.Factor_dep
    deal_time = time_grid.time_grid[deal_data.Time_dep.deal_time_grid]
    cash_start_idx = factor_dep['Cashflows'].get_cashflow_start_index(deal_time)

    # first precalc all past resets
    resets = factor_dep['Cashflows'].Resets
    known_resets = resets.known_resets(shared.simulation_batch)
    sim_resets = resets.schedule[(resets.schedule[:, utils.RESET_INDEX_Scenario] > -1) &
                                 (resets.schedule[:, utils.RESET_INDEX_Reset_Day] <=
                                  deal_time[:, utils.TIME_GRID_MTM].max())]
    all_resets = utils.calc_time_grid_curve_rate(factor_dep['ForwardPrice'], sim_resets, shared)
    all_fx_spot = utils.calc_fx_cross(factor_dep['ForwardFX'][0], factor_dep['CashFX'][0],
                                      sim_resets, shared)

    reset_values = tf.expand_dims(
        tf.squeeze(all_resets.gather_weighted_curve(
            shared, sim_resets[:, utils.RESET_INDEX_End_Day].reshape(-1, 1), multiply_by_time=False),
            axis=1) * all_fx_spot, axis=1) \
        if sim_resets.any() else tf.zeros([0, 1, shared.simulation_batch], dtype=shared.precision)

    old_resets = tf.squeeze(tf.concat([tf.stack(known_resets), reset_values]
                                      if known_resets
                                      else reset_values, axis=0), axis=1)

    forwards = utils.calc_time_grid_curve_rate(factor_dep['ForwardPrice'], deal_time, shared)
    discounts = utils.calc_time_grid_curve_rate(factor_dep['Discount'], deal_time, shared)

    start_index, start_counts = np.unique(cash_start_idx, return_counts=True)

    for index, (forward_block, discount_block) in enumerate(utils.split_counts(
            [forwards, discounts], start_counts, shared)):

        cashflows = factor_dep['Cashflows'].schedule[start_index[index]:]
        cash_offset = factor_dep['Cashflows'].offsets[start_index[index]:]

        cash_pmts, cash_index, cash_counts = np.unique(cashflows[:, utils.CASHFLOW_INDEX_Pay_Day],
                                                       return_index=True, return_counts=True)

        reset_offset = factor_dep['Cashflows'].offsets[start_index[index]][1]

        time_ofs = 0
        time_block = discount_block.time_grid[:, utils.TIME_GRID_MTM]
        future_pmts = cash_pmts.reshape(1, -1) - time_block.reshape(-1, 1)
        reset_block = resets.schedule[reset_offset:]
        reset_ofs, reset_count = np.unique(resets.split_block_resets(
            reset_offset, time_block), return_counts=True)

        # discount rates
        discount_rates = utils.calc_discount_rate(discount_block,
                                                  future_pmts, shared)

        # we need to split the forward block further
        forward_blocks = forward_block.split_counts(
            reset_count, shared) if len(reset_count) > 1 else [forward_block]

        # empty list for payments
        payments = []

        for offset, size, forward_rates in zip(*[reset_ofs, reset_count, forward_blocks]):
            # past resets
            past_resets = tf.tile(
                tf.expand_dims(old_resets[reset_offset:reset_offset + offset], axis=0), [size, 1, 1])

            # future resets
            future_ends = np.tile(reset_block[offset:, utils.RESET_INDEX_End_Day], [size, 1])

            if future_ends.any():
                future_resets = forward_rates.gather_weighted_curve(
                    shared, future_ends, multiply_by_time=False)

                forwardfx = utils.calc_fx_forward(
                    factor_dep['ForwardFX'], factor_dep['CashFX'],
                    reset_block[offset:, utils.RESET_INDEX_Start_Day],
                    discounts.time_grid[time_ofs:time_ofs + size], shared)

                all_resets = tf.concat([past_resets, future_resets * forwardfx], axis=1)
            else:
                all_resets = past_resets

            time_ofs += size

            all_payoffs = all_resets * reset_block[:, utils.RESET_INDEX_Weight].reshape(1, -1, 1)
            payoff = tf.stack(
                [tf.reduce_sum(x, axis=1) for x in tf.split(all_payoffs, cash_offset[:, 0], axis=1)], axis=1)

            # now we can price the cashflows
            payment = cashflows[:, utils.CASHFLOW_INDEX_Nominal] * (
                    cashflows[:, utils.CASHFLOW_INDEX_Start_Mult] * payoff +
                    cashflows[:, utils.CASHFLOW_INDEX_FloatMargin])

            payments.append(payment)

        # now finish the payments
        all_payments = tf.concat(payments, axis=0)

        # settle any cashflows - use this tf.gather(all_payments, time_grid.mtm_time_grid.searchsorted(cash_pmts))!!
        cash_settle(shared, factor_dep['SettleCurrency'],
                    np.searchsorted(time_grid.mtm_time_grid, cash_pmts[0]), all_payments[-1][0])
        # add it to the list
        mtm_list.append(tf.reduce_sum(all_payments * discount_rates, axis=1))

    return tf.concat(mtm_list, axis=0, name='mtm')


def pvcreditcashflows(shared, time_grid, deal_data):
    mtm_list = []
    factor_dep = deal_data.Factor_dep
    deal_time = time_grid.time_grid[deal_data.Time_dep.deal_time_grid]

    cash_start_idx = factor_dep['Cashflows'].get_cashflow_start_index(deal_time)

    discounts = utils.calc_time_grid_curve_rate(factor_dep['Discount'], deal_time, shared)
    surv = utils.calc_time_grid_curve_rate(factor_dep['Name'], deal_time, shared)

    start_index, counts = np.unique(cash_start_idx, return_counts=True)
    for index, (discount_block, surv_block) in enumerate(
            utils.split_counts([discounts, surv], counts, shared)):
        cashflows = factor_dep['Cashflows'].schedule[start_index[index]:]

        cash_pmts, cash_index = np.unique(cashflows[:, utils.CASHFLOW_INDEX_Pay_Day],
                                          return_index=True)
        cash_sts = np.unique(cashflows[:, utils.CASHFLOW_INDEX_Start_Day])

        # payment times            
        time_block = discount_block.time_grid[:, utils.TIME_GRID_MTM]
        future_pmts = cash_pmts.reshape(1, -1) - time_block.reshape(-1, 1)
        start_pmts = (cash_sts.reshape(1, -1) - time_block.reshape(-1, 1)).clip(0, np.inf)

        Dt_T = utils.calc_discount_rate(discount_block,
                                        future_pmts, shared)
        Dt_Tm1 = utils.calc_discount_rate(discount_block,
                                          start_pmts, shared)

        survival_T = utils.calc_discount_rate(surv_block, future_pmts, shared,
                                              multiply_by_time=False)
        survival_t = utils.calc_discount_rate(surv_block, start_pmts, shared,
                                              multiply_by_time=False)

        interest = cashflows[cash_index, utils.CASHFLOW_INDEX_FixedRate] * cashflows[
            cash_index, utils.CASHFLOW_INDEX_Year_Frac]
        premium = (interest[cash_index] *
                   cashflows[cash_index, utils.CASHFLOW_INDEX_Nominal]).reshape(1, -1, 1) * survival_T * Dt_T
        credit = (1.0 - factor_dep['Recovery_Rate']) * cashflows[cash_index, utils.CASHFLOW_INDEX_Nominal].reshape(
            1, -1, 1) * 0.5 * (Dt_T + Dt_Tm1) * (survival_t - survival_T)

        # settle any cashflows - TODO
        # cash_settle(shared, factor_dep['SettleCurrency'], np.searchsorted(time_grid.mtm_time_grid, cash_pmts[0]),
        #                payments[0])

        mtm_list.append(tf.reduce_sum(credit - premium, axis=1))

    return tf.concat(mtm_list, axis=0, name='mtm')


def pvequitycashflows(shared, time_grid, deal_data):
    mtm_list = []
    factor_dep = deal_data.Factor_dep
    deal_time = time_grid.time_grid[deal_data.Time_dep.deal_time_grid]
    eq_spot = utils.calc_time_grid_spot_rate(factor_dep['Equity'], deal_time, shared)
    cash, divis = factor_dep['Flows']

    # needed for grouping 
    cash_start_idx = np.searchsorted(cash.schedule[:, utils.CASHFLOW_INDEX_Start_Day],
                                     deal_time[:, utils.TIME_GRID_MTM], side='right')
    cash_end_idx = np.searchsorted(cash.schedule[:, utils.CASHFLOW_INDEX_End_Day],
                                   deal_time[:, utils.TIME_GRID_MTM], side='right')
    cash_pay_idx = cash.get_cashflow_start_index(deal_time)

    # would prefer to use np.unique here but it does not seem to work with tuples  
    all_idx = OrderedDict()
    for idx in zip(cash_start_idx, cash_end_idx, cash_pay_idx):
        all_idx[idx] = all_idx.setdefault(idx, 0) + 1

    # first precalc all past resets
    samples = cash.Resets
    known_sample = samples.known_resets(shared.simulation_batch, include_today=True)
    known_divs = samples.known_resets(shared.simulation_batch, utils.RESET_INDEX_Weight,
                                      include_today=True)
    sim_samples = samples.schedule[(samples.schedule[:, utils.RESET_INDEX_Value] == 0.0) &
                                   (samples.schedule[:, utils.RESET_INDEX_Reset_Day] <=
                                    deal_time[:, utils.TIME_GRID_MTM].max())]

    # now calculate the dividends
    div_samples = divis.Resets

    h_t0_t1 = utils.calc_realized_dividends(
        factor_dep['Equity'], factor_dep['Equity_Zero'], factor_dep['Dividend_Yield'],
        div_samples, shared, offsets=divis.offsets[:, 0])

    discounts = utils.calc_time_grid_curve_rate(factor_dep['Discount'], deal_time, shared)
    repo_discounts = utils.calc_time_grid_curve_rate(factor_dep['Equity_Zero'], deal_time, shared)
    past_samples = utils.calc_time_grid_spot_rate(factor_dep['Equity'],
                                                  sim_samples[:, :utils.RESET_INDEX_Scenario + 1],
                                                  shared)
    # fetch all fixed resets
    if past_samples.shape[1].value!=shared.simulation_batch:
        past_samples = tf.tile(past_samples, [1, shared.simulation_batch])
    
    all_samples = tf.concat(
        [tf.concat(known_sample, axis=0), past_samples], axis=0) if known_sample else past_samples

    cashflows = cash.schedule
    all_index, all_counts = zip(*all_idx.items())

    for index, (discount_block, repo_block, eq_block) in enumerate(
            utils.split_counts([discounts, repo_discounts, eq_spot], np.array(all_counts), shared)):

        start_idx, end_idx, pay_idx = all_index[index]

        cashflow_start = cashflows[start_idx:, utils.CASHFLOW_INDEX_Start_Day].reshape(1, -1)
        cashflow_pay = cashflows[pay_idx:, utils.CASHFLOW_INDEX_Pay_Day].reshape(1, -1)

        time_block = discount_block.time_grid[:, utils.TIME_GRID_MTM]
        future_pmts = cashflow_pay - time_block.reshape(-1, 1)
        payoffs = []
        discount_rates = utils.calc_discount_rate(discount_block,
                                                  future_pmts, shared)

        # need equity forwards for start and end cashflows
        if pay_idx < end_idx:
            future_end = cashflows[pay_idx, utils.CASHFLOW_INDEX_End_Day].reshape(1, -1)

            St0 = all_samples[2 * pay_idx]
            St1 = all_samples[2 * pay_idx + 1]
            Ht0_t1 = 0 * h_t0_t1[pay_idx]

            payoff = (cashflows[pay_idx, utils.CASHFLOW_INDEX_End_Mult].reshape(1, -1, 1) * St1 -
                      cashflows[pay_idx, utils.CASHFLOW_INDEX_Start_Mult].reshape(1, -1, 1) * St0 +
                      cashflows[pay_idx, utils.CASHFLOW_INDEX_Dividend_Mult].reshape(1, -1, 1) * Ht0_t1)
            payment = payoff * cashflows[pay_idx, utils.CASHFLOW_INDEX_FixedAmt].reshape(1, -1, 1)

            if factor_dep['PrincipleNotShares']:
                payment /= St0

            payoffs.append(payment)
            # settle cashflow if necessary
            cash_settle(shared, factor_dep['SettleCurrency'],
                        np.searchsorted(time_grid.mtm_time_grid, cashflow_pay[0][0]),
                        payment[0][0])

        if end_idx < start_idx:
            cashflow_end = cashflows[end_idx, utils.CASHFLOW_INDEX_End_Day].reshape(1, -1)
            future_end = cashflow_end - time_block.reshape(-1, 1)
            forward_end = utils.calc_eq_forward(factor_dep['Equity'], factor_dep['Equity_Zero'],
                                                factor_dep['Dividend_Yield'],
                                                np.squeeze(cashflow_end, axis=0),
                                                discount_block.time_grid, shared)
            discount_end = utils.calc_discount_rate(
                repo_block, future_end, shared)

            St0 = all_samples[end_idx * 2]
            Ht0_t = utils.calc_realized_dividends(
                factor_dep['Equity'], factor_dep['Equity_Zero'], factor_dep['Dividend_Yield'],
                utils.calc_dividend_samples(time_block[0], time_block[-1], time_grid), shared)
            payoff = ((cashflows[end_idx, utils.CASHFLOW_INDEX_End_Mult] -
                       cashflows[end_idx, utils.CASHFLOW_INDEX_Dividend_Mult]) * forward_end +
                      cashflows[end_idx, utils.CASHFLOW_INDEX_Dividend_Mult] * (
                          tf.expand_dims(eq_block + Ht0_t, axis=1)) / discount_end -
                      cashflows[end_idx, utils.CASHFLOW_INDEX_Start_Mult] * St0)
            if factor_dep['PrincipleNotShares']:
                payoff /= tf.reshape(St0, [1, 1, -1])

            payoffs.append(payoff * cashflows[end_idx, utils.CASHFLOW_INDEX_FixedAmt].reshape(1, -1, 1))

        if cashflow_start.any():
            cashflow_end = cashflows[start_idx:, utils.CASHFLOW_INDEX_End_Day].reshape(1, -1)
            future_start = cashflow_start - time_block.reshape(-1, 1)
            future_end = cashflow_end - time_block.reshape(-1, 1)
            forward_start = utils.calc_eq_forward(factor_dep['Equity'], factor_dep['Equity_Zero'],
                                                  factor_dep['Dividend_Yield'], np.squeeze(cashflow_start, axis=0),
                                                  discount_block.time_grid, shared)
            forward_end = utils.calc_eq_forward(factor_dep['Equity'], factor_dep['Equity_Zero'],
                                                factor_dep['Dividend_Yield'], np.squeeze(cashflow_end, axis=0),
                                                discount_block.time_grid, shared)
            discount_start = utils.calc_discount_rate(
                repo_block, future_start, shared)
            discount_end = utils.calc_discount_rate(
                repo_block, future_end, shared)

            if factor_dep['PrincipleNotShares']:
                factor1 = forward_end / forward_start
                factor2 = 1.0
            else:
                factor1 = forward_end
                factor2 = forward_start

            payoff = (cashflows[start_idx:, utils.CASHFLOW_INDEX_End_Mult] -
                      cashflows[start_idx:, utils.CASHFLOW_INDEX_Dividend_Mult]).reshape(1, -1, 1) * factor1 + (
                             cashflows[start_idx:, utils.CASHFLOW_INDEX_Dividend_Mult].reshape(1, -1, 1) *
                             (discount_start / discount_end) -
                             cashflows[start_idx:, utils.CASHFLOW_INDEX_Start_Mult].reshape(1, -1, 1)) * factor2

            payoffs.append(payoff * cashflows[start_idx:, utils.CASHFLOW_INDEX_FixedAmt].reshape(1, -1, 1))

        # now finish the payments
        payments = tf.concat(payoffs, axis=1) if len(payoffs) > 1 else payoffs[0]
        mtm_list.append(tf.reduce_sum(payments * discount_rates, axis=1))

    return tf.concat(mtm_list, axis=0, name='mtm')


def pvfixedleg(shared, time_grid, deal_data):
    deal_time = time_grid.time_grid[deal_data.Time_dep.deal_time_grid]
    FX_rep = utils.calc_fx_cross(deal_data.Factor_dep['Currency'][0], shared.Report_Currency,
                                 deal_time, shared)

    mtm = FX_rep * pvfixedcashflows(shared, time_grid, deal_data)

    return mtm


def pvenergyleg(shared, time_grid, deal_data):
    deal_time = time_grid.time_grid[deal_data.Time_dep.deal_time_grid]
    FX_rep = utils.calc_fx_cross(deal_data.Factor_dep['Payoff_Currency'], shared.Report_Currency,
                                 deal_time, shared)

    mtm = FX_rep * pvenergycashflows(shared, time_grid, deal_data)

    return mtm


def pvfloatleg(shared, time_grid, deal_data):
    deal_time = time_grid.time_grid[deal_data.Time_dep.deal_time_grid]
    FX_rep = utils.calc_fx_cross(deal_data.Factor_dep['Currency'][0], shared.Report_Currency,
                                 deal_time, shared)

    mtm = FX_rep * pvfloatcashflowlist(shared, time_grid, deal_data, pricer_float_cashflows)

    return mtm


def pvcapleg(shared, time_grid, deal_data):
    deal_time = time_grid.time_grid[deal_data.Time_dep.deal_time_grid]
    FX_rep = utils.calc_fx_cross(deal_data.Factor_dep['Currency'][0], shared.Report_Currency,
                                 deal_time, shared)
    return FX_rep * pvfloatcashflowlist(shared, time_grid, deal_data, pricer_cap)


def pvfloorleg(shared, time_grid, deal_data):
    deal_time = time_grid.time_grid[deal_data.Time_dep.deal_time_grid]
    FX_rep = utils.calc_fx_cross(deal_data.Factor_dep['Currency'][0], shared.Report_Currency,
                                 deal_time, shared)
    return FX_rep * pvfloatcashflowlist(shared, time_grid, deal_data, pricer_floor)


def pvindexleg(shared, time_grid, deal_data):
    deal_time = time_grid.time_grid[deal_data.Time_dep.deal_time_grid]
    FX_rep = utils.calc_fx_cross(deal_data.Factor_dep['Currency'], shared.Report_Currency,
                                 deal_time, shared)

    mtm = FX_rep * pvindexcashflows(shared, time_grid, deal_data)

    return mtm


def pvcdsleg(shared, time_grid, deal_data):
    deal_time = time_grid.time_grid[deal_data.Time_dep.deal_time_grid]
    FX_rep = utils.calc_fx_cross(deal_data.Factor_dep['Currency'], shared.Report_Currency,
                                 deal_time, shared)
    mtm = FX_rep * pvcreditcashflows(shared, time_grid, deal_data)

    return mtm


def pvequityleg(shared, time_grid, deal_data):
    deal_time = time_grid.time_grid[deal_data.Time_dep.deal_time_grid]
    FX_rep = utils.calc_fx_cross(deal_data.Factor_dep['Currency'], shared.Report_Currency,
                                 deal_time, shared)

    mtm = FX_rep * pvequitycashflows(shared, time_grid, deal_data)

    return mtm


def pvfxbarrieroption(shared, time_grid, deal_data):
    deal_time = time_grid.time_grid[deal_data.Time_dep.deal_time_grid]
    nominal = deal_data.Instrument.field['Underlying_Amount']
    payoff_currency = deal_data.Instrument.field[deal_data.Instrument.field['Payoff_Currency']]

    curr_curve = utils.calc_time_grid_curve_rate(
        deal_data.Factor_dep['Currency'][1], deal_time[:-1], shared)
    und_curr_curve = utils.calc_time_grid_curve_rate(
        deal_data.Factor_dep['Underlying_Currency'][1], deal_time[:-1], shared)

    spot = utils.calc_fx_cross(deal_data.Factor_dep['Underlying_Currency'][0],
                               deal_data.Factor_dep['Currency'][0], deal_time, shared)

    # need to adjust if there's just 1 timepoint - i.e. base reval
    if time_grid.mtm_time_grid.size > 1:
        tau = (deal_data.Factor_dep['Expiry'] - deal_time[:, utils.TIME_GRID_MTM])[:-1]
    else:
        tau = (deal_data.Factor_dep['Expiry'] - deal_time[:, utils.TIME_GRID_MTM])

    b = tf.squeeze(curr_curve.gather_weighted_curve(shared, tau.reshape(-1, 1), multiply_by_time=False) -
                   und_curr_curve.gather_weighted_curve(shared, tau.reshape(-1, 1), multiply_by_time=False), axis=1)

    fx_rep = utils.calc_fx_cross(deal_data.Factor_dep['Currency'][0], shared.Report_Currency,
                                 deal_time, shared)

    pv = pvbarrieroption(
        shared, time_grid, deal_data, nominal, spot, b, tau, payoff_currency,
        invert_moneyness=deal_data.Factor_dep['Invert_Moneyness'])

    mtm = fx_rep * pv

    return mtm


def pveqbarrieroption(shared, time_grid, deal_data):
    deal_time = time_grid.time_grid[deal_data.Time_dep.deal_time_grid]
    nominal = deal_data.Instrument.field['Units']
    payoff_currency = deal_data.Instrument.field[deal_data.Instrument.field['Currency']]

    eq_zer_curve = utils.calc_time_grid_curve_rate(
        deal_data.Factor_dep['Equity_Zero'], deal_time[:-1], shared)
    eq_div_curve = utils.calc_time_grid_curve_rate(
        deal_data.Factor_dep['Dividend_Yield'], deal_time[:-1], shared)

    spot = utils.calc_time_grid_spot_rate(deal_data.Factor_dep['Equity'], deal_time, shared)

    # need to adjust if there's just 1 timepoint - i.e. base reval
    if time_grid.mtm_time_grid.size > 1:
        tau = (deal_data.Factor_dep['Expiry'] - deal_time[:, utils.TIME_GRID_MTM])[:-1]
    else:
        tau = (deal_data.Factor_dep['Expiry'] - deal_time[:, utils.TIME_GRID_MTM])

    b = tf.squeeze(eq_zer_curve.gather_weighted_curve(shared, tau.reshape(-1, 1),
                                               multiply_by_time=False) -
                   eq_div_curve.gather_weighted_curve(shared, tau.reshape(-1, 1),
                                               multiply_by_time=False), axis=1)

    fx_rep = utils.calc_fx_cross(deal_data.Factor_dep['Currency'], shared.Report_Currency,
                                 deal_time, shared)

    pv = pvbarrieroption(
        shared, time_grid, deal_data, nominal, spot, b, tau, payoff_currency,
        invert_moneyness=deal_data.Factor_dep['Invert_Moneyness'])

    mtm = fx_rep * pv

    return mtm


def pvfxonetouch(shared, time_grid, deal_data):
    deal_time = time_grid.time_grid[deal_data.Time_dep.deal_time_grid]
    nominal = deal_data.Instrument.field['Cash_Payoff']
    payoff_currency = deal_data.Instrument.field[deal_data.Instrument.field['Payoff_Currency']]
    fx_rep = utils.calc_fx_cross(deal_data.Factor_dep['Currency'][0],
                                 shared.Report_Currency, deal_time, shared)

    curr_curve = utils.calc_time_grid_curve_rate(
        deal_data.Factor_dep['Currency'][1], deal_time[:-1], shared)
    und_curr_curve = utils.calc_time_grid_curve_rate(
        deal_data.Factor_dep['Underlying_Currency'][1], deal_time[:-1], shared)

    spot = utils.calc_fx_cross(deal_data.Factor_dep['Underlying_Currency'][0],
                               deal_data.Factor_dep['Currency'][0], deal_time, shared)

    # need to adjust if there's just 1 timepoint - i.e. base reval
    if time_grid.mtm_time_grid.size > 1:
        tau = (deal_data.Factor_dep['Expiry'] - deal_time[:, utils.TIME_GRID_MTM])[:-1]
    else:
        tau = (deal_data.Factor_dep['Expiry'] - deal_time[:, utils.TIME_GRID_MTM])

    b = tf.squeeze(curr_curve.gather_weighted_curve(shared, tau.reshape(-1, 1), multiply_by_time=False) -
                   und_curr_curve.gather_weighted_curve(shared, tau.reshape(-1, 1), multiply_by_time=False), axis=1)

    pv = pvonetouchoption(
        shared, time_grid, deal_data, nominal,
        spot, b, tau, payoff_currency, invert_moneyness=deal_data.Factor_dep['Invert_Moneyness'])

    mtm = fx_rep * pv

    return mtm



#########################################################################
# Copyright (C) 2016-2019  Shuaib Osman (sosman@investec.co.za)
# This file is part of RiskFlow.
#
# RiskFlow is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# RiskFlow is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with RiskFlow.  If not, see <http://www.gnu.org/licenses/>.
#########################################################################

"""
All Bootstrappers are defined here.
"""

# import standard libraries
import time
import logging
from collections import OrderedDict, namedtuple

# third party stuff
import hdsobol
import numpy as np
import pandas as pd
import tensorflow as tf
import scipy.interpolate

# Internal modules
import utils
import pricing
import instruments
import riskfactors
import stochasticprocess

# misc functions/classes
from calculation import TimeGrid
from tensorflow.contrib.opt import ExternalOptimizerInterface

from tensorflow.python.framework import ops
from tensorflow.python.ops import array_ops
from tensorflow.python.platform import tf_logging as logging

from scipy.stats import norm
from tensorflow.python.client import device_lib

curve_calibration_class = namedtuple('shared_mem', 't_Buffer \
                        t_Scenario_Buffer precision simulation_batch')

curve_jacobian_class = namedtuple('shared_mem', 't_Buffer t_Static_Buffer \
                                precision riskneutral simulation_batch')

market_swap_class = namedtuple('market_swap', 'deal_data price weight')
date_desc = {'years': 'Y', 'months': 'M', 'days': 'D'}
# date formatter
date_fmt = lambda x: ''.join(['{0}{1}'.format(v, date_desc[k]) for k, v in x.kwds.items()])


def create_float_cashflows(base_date, cashflow_obj, frequency):
    cashflows = []
    for cashflow, reset in zip(cashflow_obj.schedule, cashflow_obj.Resets.schedule):
        cashflows.append(OrderedDict([
            ('Payment_Date', base_date + pd.offsets.Day(cashflow[utils.CASHFLOW_INDEX_Pay_Day])),
            ('Notional', 1.0),
            ('Accrual_Start_Date', base_date + pd.offsets.Day(cashflow[utils.CASHFLOW_INDEX_Start_Day])),
            ('Accrual_End_Date', base_date + pd.offsets.Day(cashflow[utils.CASHFLOW_INDEX_End_Day])),
            ('Accrual_Year_Fraction', cashflow[utils.CASHFLOW_INDEX_Year_Frac]),
            ('Fixed_Amount', cashflow[utils.CASHFLOW_INDEX_FixedAmt]),
            ('Resets', [[base_date + pd.offsets.Day(reset[utils.RESET_INDEX_Reset_Day]),
                         base_date + pd.offsets.Day(reset[utils.RESET_INDEX_Start_Day]),
                         base_date + pd.offsets.Day(reset[utils.RESET_INDEX_End_Day]),
                         reset[utils.RESET_INDEX_Accrual],
                         frequency, 'ACT_365', '0D', 0.0, 'No', utils.Percent(0.0)]]),
            ('Margin', utils.Basis(0.0))
        ]))
    return cashflows


def normalize(sample):
    '''Simple function to ensure that the sample used for the monte carlo sim has mean 0 and var 1'''
    return (sample - sample.mean(axis=0)) / sample.std(axis=0)


def atm_swap(base_date, curve_factor, time_grid, effective, maturity, float_freq, daycount):
    float_pay_dates = instruments.generate_dates_backward(
        maturity, effective, float_freq)

    float_cash = utils.generate_float_cashflows(
        base_date, time_grid, float_pay_dates, 1.0, None, None,
        float_freq, pd.DateOffset(months=0), utils.get_day_count(daycount), 0.0)

    K, pvbp = float_cash.get_par_swap_rate(base_date, curve_factor)
    float_cash.set_fixed_amount(-K)
    return float_cash, K, pvbp


def atm_depo(base_date, curve_factor, time_grid, maturity, daycount):
    time_in_years = utils.get_day_count_accrual(
        base_date, (maturity - base_date).days, utils.get_day_count(daycount))

    fixed_pay_dates = instruments.generate_dates_backward(
        maturity, base_date, maturity - base_date)

    fixed_cash = utils.generate_fixed_cashflows(
        base_date, fixed_pay_dates, 1.0, None,
        utils.get_day_count(daycount), 0.0)

    fixed_cash.overwrite_rate(utils.CASHFLOW_INDEX_FixedAmt, 1.0)

    return fixed_cash


def create_market_swaps(base_date, time_grid, curve_index, vol_surface, curve_factor,
                        instrument_definitions, rate=None):
    # store these benchmark swap definitions if necessary
    benchmarks = []
    # store the benchmark instruments
    all_deals = {}
    # cater for shifted lognormal vols
    shift_parameter = vol_surface.BlackScholesDisplacedShiftValue / 100.0
    for instrument in instrument_definitions:
        # setup the instrument
        effective = base_date + instrument['Start']
        maturity = effective + instrument['Tenor']
        exp_days = (effective - base_date).days
        tenor = (maturity - effective).days / utils.DAYS_IN_YEAR
        expiry = exp_days / utils.DAYS_IN_YEAR
        time_index = np.searchsorted(time_grid.mtm_time_grid, [exp_days], side='right') - 1
        swaption_name = 'Swap_{0:05.2f}_{1:02.0f}_{2}_{3}'.format(
            expiry, tenor, date_fmt(instrument['Start']), date_fmt(instrument['Tenor']))

        float_pay_dates = instruments.generate_dates_backward(
            maturity, effective, instrument['Floating_Frequency'])

        float_cash = utils.generate_float_cashflows(
            base_date, time_grid, float_pay_dates, 1.0, None, None,
            instrument['Floating_Frequency'], pd.DateOffset(month=0),
            utils.get_day_count(instrument['Day_Count']), 0.0)

        K, pvbp = float_cash.get_par_swap_rate(base_date, curve_factor)

        if instrument['Fixed_Frequency'] != instrument['Floating_Frequency']:
            fixed_pay_dates = instruments.generate_dates_backward(
                maturity, effective, instrument['Fixed_Frequency'])
            fixed_cash = utils.generate_fixed_cashflows(
                base_date, fixed_pay_dates, 1.0, None, utils.get_day_count(instrument['Day_Count']), 0.0)
            pv_float = K * pvbp
            pvbp = fixed_cash.get_par_swap_rate(base_date, curve_factor)
            K = pv_float / pvbp
            fixed_cash.set_fixed_amount(K)
            fixed_indices = float_cash[:, utils.CASHFLOW_INDEX_Pay_Day].searchsorted(
                fixed_cash[:, utils.CASHFLOW_INDEX_Pay_Day])

            if not (float_cash[fixed_indices, utils.CASHFLOW_INDEX_Pay_Day] ==
                    fixed_cash[:, utils.CASHFLOW_INDEX_Pay_Day]).all():
                raise Exception('Float leg and Fixed legs do not coincide')

            # set the float leg fixed amount
            float_cash.schedule[fixed_indices, utils.CASHFLOW_INDEX_FixedAmt] =\
                -fixed_cash[:, utils.CASHFLOW_INDEX_FixedAmt]
        else:
            float_cash.set_fixed_amount(-K)

        # get the atm vol
        if instrument['Market_Volatility'].amount:
            vol = instrument['Market_Volatility'].amount
        else:
            vol = vol_surface.ATM(tenor, expiry)[0][0]

        deal_data = utils.DealDataType(
            Instrument=None, Factor_dep={'Cashflows': float_cash, 'Forward': curve_index,
                                         'Discount': curve_index, 'CompoundingMethod': 'None'},
            Time_dep=utils.DealTimeDependencies(time_grid.mtm_time_grid, time_index), Calc_res=None)

        shifted_strike = K + shift_parameter
        # store this
        all_deals[swaption_name] = market_swap_class(
            deal_data=deal_data,
            price=pvbp * utils.black_european_option_price(shifted_strike, shifted_strike, 0.0, vol, expiry, 1.0, 1.0),
            weight=instrument['Weight'])

        # store the benchmark
        if rate is not None:
            benchmarks.append(
                instruments.construct_instrument(
                    {'Object': 'CFFloatingInterestListDeal',
                     'Reference': swaption_name,
                     'Currency': curve_factor.param['Currency'],
                     'Discount_Rate': '.'.join(rate),
                     'Forecast_Rate': '.'.join(rate),
                     'Buy_Sell': 'Buy',
                     'Cashflows': {'Items': create_float_cashflows(
                         base_date, float_cash, instrument['Floating_Frequency'])}},
                    {})
            )

    return all_deals, benchmarks


def _get_shape_tuple(tensor):
    return tuple(dim.value for dim in tensor.get_shape())


def _prod(array):
    prod = 1
    for value in array:
        prod *= value
    return prod


def _accumulate(list_):
    total = 0
    yield total
    for x in list_:
        total += x
        yield total


class ScipyLeastsqOptimizerInterface(object):
    """Base class for interfaces with external optimization algorithms.

    Subclass this and implement `_minimize` in order to wrap a new optimization
    algorithm.

    `ExternalOptimizerInterface` should not be instantiated directly; instead use
    e.g. `ScipyOptimizerInterface`.

    @@__init__

    @@minimize
    """

    def __init__(self, loss, var_to_bounds, var_list, **optimizer_kwargs):

        def fwd_gradients(ys, xs, d_xs):
            """ Forward-mode pushforward analogous to the pullback defined by tf.gradients.
                With tf.gradients, grad_ys is the vector being pulled back, and here d_xs is
                the vector being pushed forward."""

            v = self._dummy_ys_vec  # dummy variable
            g = tf.concat([x for x in tf.gradients(ys, xs, grad_ys=v) if x is not None], axis=0)
            return tf.gradients(g, v, grad_ys=d_xs)

        self._loss = tf.stack(loss)
        self._vars = var_list
        packed_bounds = None
        if var_to_bounds is not None:
            left_packed_bounds = []
            right_packed_bounds = []
            for var in self._vars:
                shape = var.get_shape().as_list()
                bounds = (-np.infty, np.infty)
                if var in var_to_bounds:
                    bounds = var_to_bounds[var]
                    left_packed_bounds.extend(list(np.broadcast_to(bounds[0], shape).flat))
                    right_packed_bounds.extend(list(np.broadcast_to(bounds[1], shape).flat))
            packed_bounds = list([left_packed_bounds, right_packed_bounds])
        self._packed_bounds = packed_bounds

        self._dummy_ys_vec = tf.placeholder_with_default(
            np.ones(self._loss.get_shape()[0].value, dtype=np.float32), shape=self._loss.get_shape())
        self._dummy_jac_vec = array_ops.placeholder(
            self._vars[0].dtype, sum([_get_shape_tuple(x)[0] for x in self._vars]))

        self._update_placeholders = [
            array_ops.placeholder(var.dtype) for var in self._vars
        ]
        self._var_updates = [
            var.assign(array_ops.reshape(placeholder, _get_shape_tuple(var)))
            for var, placeholder in zip(self._vars, self._update_placeholders)
        ]

        time_now = time.clock()
        logging.info('Calculating jacobian')
        # loss_grads = [tf.gradients(resid, self._vars) for resid in loss]
        self.jvp = fwd_gradients(self._loss, self._vars, self._dummy_jac_vec)
        logging.info('Done Calculating jacobian')
        # record the time
        logging.info('This took {} seconds'.format(time.clock() - time_now))

        self.optimizer_kwargs = optimizer_kwargs

        self._packed_var = self._pack(self._vars)
        self._packed_equality_grads = []
        self._packed_inequality_grads = []
        dims = [_prod(_get_shape_tuple(var)) for var in self._vars]
        accumulated_dims = list(_accumulate(dims))
        self._packing_slices = [
            slice(start, end)
            for start, end in zip(accumulated_dims[:-1], accumulated_dims[1:])
        ]

    def minimize(self, session=None, feed_dict=None, fetches=None,
                 step_callback=None, loss_callback=None, **run_kwargs):
        """Minimize a scalar `Tensor`.

        Variables subject to optimization are updated in-place at the end of
        optimization.

        Note that this method does *not* just return a minimization `Op`, unlike
        `Optimizer.minimize()`; instead it actually performs minimization by
        executing commands to control a `Session`.

        Args:
          session: A `Session` instance.
          feed_dict: A feed dict to be passed to calls to `session.run`.
          fetches: A list of `Tensor`s to fetch and supply to `loss_callback`
            as positional arguments.
          step_callback: A function to be called at each optimization step;
            arguments are the current values of all optimization variables
            flattened into a single vector.
          loss_callback: A function to be called every time the loss and gradients
            are computed, with evaluated fetches supplied as positional arguments.
          **run_kwargs: kwargs to pass to `session.run`.
        """

        session = session or ops.get_default_session()
        feed_dict = feed_dict or {}
        fetches = fetches or []

        loss_callback = loss_callback or (lambda *fetches: None)
        step_callback = step_callback or (lambda xk: None)

        # Construct loss function and associated gradient.
        loss_func = self._make_eval_func(self._loss, session,
                                         feed_dict, fetches, loss_callback)
        n = self._dummy_jac_vec.shape[0].value
        grad_func = [self._make_eval_func(self.jvp, session,
                                          {self._dummy_jac_vec: np.eye(n)[i]},
                                          fetches, loss_callback)
                     for i in range(n)]

        # Get initial value from TF session.
        initial_packed_var_val = session.run(self._packed_var)

        # Perform minimization.
        packed_var_val = self._minimize(
            initial_val=initial_packed_var_val,
            loss_func=loss_func, grad_func=grad_func,
            packed_bounds=self._packed_bounds,
            step_callback=step_callback,
            optimizer_kwargs=self.optimizer_kwargs)
        var_vals = [
            packed_var_val[packing_slice] for packing_slice in self._packing_slices
        ]

        # Set optimization variables to their new values.
        session.run(
            self._var_updates,
            feed_dict=dict(zip(self._update_placeholders, var_vals)),
            **run_kwargs)

    def _minimize(self, initial_val, loss_func, grad_func,
                  packed_bounds, step_callback, optimizer_kwargs):

        def loss_func_wrapper(x):
            return loss_func(x)[0]

        def grad_func_wrapper(x):
            return np.array([grad(x)[0].astype('float64') for grad in grad_func]).T

        import scipy.optimize
        if packed_bounds is not None:
            result = scipy.optimize.least_squares(loss_func_wrapper, initial_val,
                                                  jac=grad_func_wrapper, bounds=packed_bounds)
        else:
            result = scipy.optimize.least_squares(loss_func_wrapper, initial_val,
                                                  jac=grad_func_wrapper)

        message_lines = [
            'Optimization terminated with:',
            '  Message: {}'.format(result.message),
            '  Objective function value: {}'.format(result.fun),
        ]

        if hasattr(result, 'nit'):
            # Some optimization methods might not provide information such as nit and
            # nfev in the return. Logs only available information.
            message_lines.append('  Number of iterations: {}'.format(result.nit))
        if hasattr(result, 'nfev'):
            message_lines.append('  Number of functions evaluations: {}'.format(result.nfev))
        logging.debug('\n'.join(message_lines))

        return result['x']

    @classmethod
    def _pack(cls, tensors):
        """Pack a list of `Tensor`s into a single, flattened, rank-1 `Tensor`."""
        if not tensors:
            return None
        elif len(tensors) == 1:
            return array_ops.reshape(tensors[0], [-1])
        else:
            flattened = [array_ops.reshape(tensor, [-1]) for tensor in tensors]
            return array_ops.concat(flattened, 0)

    def _make_eval_func(self, tensors, session, feed_dict, fetches,
                        callback=None):
        """Construct a function that evaluates a `Tensor` or list of `Tensor`s."""
        if not isinstance(tensors, list):
            tensors = [tensors]
        num_tensors = len(tensors)

        def eval_func(x):
            """Function to evaluate a `Tensor`."""
            augmented_feed_dict = {
                var: x[packing_slice].reshape(_get_shape_tuple(var))
                for var, packing_slice in zip(self._vars, self._packing_slices)
            }

            augmented_feed_dict.update(feed_dict)
            augmented_fetches = tensors + fetches

            augmented_fetch_vals = session.run(augmented_fetches,
                                               feed_dict=augmented_feed_dict)

            if callable(callback):
                callback(*augmented_fetch_vals[num_tensors:])

            return augmented_fetch_vals[:num_tensors]

        return eval_func

    def _make_eval_funcs(self, tensors, session, feed_dict, fetches, callback=None):
        return [self._make_eval_func(tensor, session, feed_dict, fetches, callback)
                for tensor in tensors]


class ScipyBasinOptimizerInterface(ExternalOptimizerInterface):
    """Wrapper allowing `scipy.optimize.basinhopping` to operate a `tf.Session`.
    Implemented exactly the same as ScipyOptimizerInterface
    """

    _DEFAULT_METHOD = 'L-BFGS-B'

    def _minimize(self, initial_val, loss_grad_func, equality_funcs,
                  equality_grad_funcs, inequality_funcs, inequality_grad_funcs,
                  packed_bounds, step_callback, optimizer_kwargs):

        def loss_grad_func_wrapper(x):
            # SciPy's L-BFGS-B Fortran implementation requires gradients as doubles.
            loss, gradient = loss_grad_func(x)
            return loss, gradient.astype('float64')

        def print_fun(x, f, accepted):
            logging.debug("at minimum %.4f accepted %d" % (f, int(accepted)))
            if f < 0.1:
                return True

        optimizer_kwargs = dict(optimizer_kwargs.items())
        method = optimizer_kwargs.pop('method', self._DEFAULT_METHOD)
        take_step = optimizer_kwargs.pop('take_step', None)
        accept_test = optimizer_kwargs.pop('accept_test', None)

        minimize_args = [loss_grad_func_wrapper, initial_val]
        minimize_kwargs = {
            'jac': True,
            'callback': step_callback,
            'method': method,
            'constraints': [],
            'bounds': packed_bounds,
        }

        import scipy.optimize
        result = scipy.optimize.basinhopping(
            *minimize_args, minimizer_kwargs=minimize_kwargs, niter=100, T=2.5,
            accept_test=accept_test, callback=print_fun, take_step=take_step)

        message_lines = [
            'Optimization terminated with:',
            '  Message: {}'.format(result.message),
            '  Objective function value: {}'.format(result.fun),
        ]

        if hasattr(result, 'nit'):
            # Some optimization methods might not provide information such as nit and
            # nfev in the return. Logs only available information.
            message_lines.append('  Number of iterations: {}'.format(result.nit))
        if hasattr(result, 'nfev'):
            message_lines.append('  Number of functions evaluations: {}'.format(result.nfev))
        logging.debug('\n'.join(message_lines))

        return result['x']


class InterestRateJacobian(object):
    def __init__(self, param, prec=np.float32):
        self.prec = prec
        self.param = param
        self.batch_size = 1

    def bootstrap(self, sys_params, price_models, price_factors, market_prices, calendars, debug=None):

        def fwd_gradients(ys, xs, d_xs):
            """ Forward-mode pushforward analogous to the pullback defined by tf.gradients.
                With tf.gradients, grad_ys is the vector being pulled back, and here d_xs is
                the vector being pushed forward."""
            v = tf.placeholder_with_default(
                np.ones([x.value for x in ys.get_shape()], dtype=self.prec), shape=ys.get_shape())
            g = tf.concat(tf.gradients(ys, xs, grad_ys=v), axis=0)
            return tf.gradients(g, v, grad_ys=d_xs)

        base_date = sys_params['Base_Date']
        # do all of this at time 0        
        time_grid = TimeGrid({base_date}, {base_date}, {base_date})
        time_grid.set_base_date(base_date)
        # now prepare for inverse bootstrap
        for market_price, implied_params in market_prices.items():
            rate = utils.check_rate_name(market_price)
            market_factor = utils.Factor(rate[0], rate[1:])
            if market_factor.type == 'InterestRatePrices':
                # get the currency 
                curr = implied_params['instrument']['Currency']
                graph = tf.Graph()
                with graph.as_default():
                    # this shouldn't fail - if it does, need to log it and move on
                    try:
                        ir_factor = utils.Factor('InterestRate', rate[1:])
                        ir_curve = riskfactors.construct_factor(ir_factor, price_factors)
                    except Exception:
                        logging.warning('Unable to calculate the Jacobian for {0} - skipping'.format(market_price))
                        continue

                    # calculate a reverse lookup for the tenors and store the daycount code
                    all_tenors = utils.update_tenors(base_date, {ir_factor: ir_curve})
                    # calculate the curve index - need to clean this up - TODO!!!
                    curve_index = [instruments.calc_factor_index(ir_factor, {ir_factor: 0}, {}, all_tenors)]
                    benchmarks = OrderedDict()
                    # state for all calcs
                    shared_mem = curve_jacobian_class(
                        t_Buffer={},
                        t_Static_Buffer=[tf.constant(ir_curve.current_value(), dtype=self.prec)],
                        precision=self.prec,
                        riskneutral=False,
                        simulation_batch=self.batch_size)

                    for child in implied_params['Children']:
                        if child['quote']['DealType'] == 'DepositDeal':
                            maturity = base_date + child['instrument']['Payment_Frequency']
                            fixed_cash = atm_depo(base_date, ir_curve, time_grid, maturity,
                                                  child['instrument']['Accrual_Day_Count'])
                            deal_data = utils.DealDataType(
                                Instrument=None, Factor_dep={'Cashflows': fixed_cash, 'Discount': curve_index},
                                Time_dep=utils.DealTimeDependencies(time_grid.mtm_time_grid, [0]), Calc_res=None)
                            fmt_fn = str if isinstance(child['quote']['Descriptor'], str) else date_fmt
                            benchmarks['Deposit_' + fmt_fn(child['quote']['Descriptor'])] = pricing.pvfixedcashflows(
                                shared_mem, time_grid, deal_data, settle_cash=False)
                        elif child['quote']['DealType'] == 'FRADeal':
                            float_cash, K, pvbp = atm_swap(
                                base_date, ir_curve, time_grid,
                                base_date + pd.DateOffset(months=child['quote']['Descriptor'].data[0]),
                                base_date + pd.DateOffset(months=child['quote']['Descriptor'].data[1]),
                                pd.DateOffset(months=np.diff(child['quote']['Descriptor'].data)[0]),
                                child['instrument']['Day_Count'])
                            deal_data = utils.DealDataType(
                                Instrument=None, Factor_dep={'Cashflows': float_cash, 'Forward': curve_index,
                                                             'Discount': curve_index, 'CompoundingMethod': 'None'},
                                Time_dep=utils.DealTimeDependencies(time_grid.mtm_time_grid, [0]), Calc_res=None)
                            benchmarks['FRA_' + str(child['quote']['Descriptor'])] = pricing.pvfloatcashflowlist(
                                shared_mem, time_grid, deal_data, pricing.pricer_float_cashflows, settle_cash=False)
                        elif child['quote']['DealType'] == 'SwapInterestDeal':
                            float_cash, K, pvbp = atm_swap(
                                base_date, ir_curve, time_grid,
                                base_date,
                                base_date + child['quote']['Descriptor'],
                                child['instrument']['Pay_Frequency'],
                                child['instrument']['Pay_Day_Count'])
                            deal_data = utils.DealDataType(
                                Instrument=None, Factor_dep={'Cashflows': float_cash, 'Forward': curve_index,
                                                             'Discount': curve_index, 'CompoundingMethod': 'None'},
                                Time_dep=utils.DealTimeDependencies(time_grid.mtm_time_grid, [0]), Calc_res=None)
                            benchmarks['Swap_' + date_fmt(child['quote']['Descriptor'])] = pricing.pvfloatcashflowlist(
                                shared_mem, time_grid, deal_data, pricing.pricer_float_cashflows, settle_cash=False)
                        else:
                            raise Exception('quote type not supported')

                    n = ir_curve.tenors.shape[0]
                    dummy_jac_vec = tf.placeholder(self.prec, n)
                    bench = tf.squeeze(tf.concat(list(benchmarks.values()), axis=0), axis=1)
                    jvp = fwd_gradients(bench, shared_mem.t_Static_Buffer, dummy_jac_vec)

                config = tf.ConfigProto(allow_soft_placement=True)

                with tf.Session(graph=graph, config=config) as sess:
                    I = np.eye(n)
                    # store the output
                    price_param = utils.Factor('InterestRateJacobian', market_factor.name)
                    jac = OrderedDict()
                    j = np.array([sess.run(jvp, {dummy_jac_vec: I[i]})[0] for i in range(n)]).T
                    for index, benchmark_name in enumerate(benchmarks.keys()):
                        non_zero = np.where(j[index] != 0.0)
                        jac[benchmark_name] = utils.Curve([], list(zip(ir_curve.tenors[non_zero], j[index][non_zero])))
                    # jac = pd.DataFrame(j, index=list(benchmarks.keys()), columns=ir_curve.tenors)

                price_factors[utils.check_tuple_name(price_param)] = jac


class GBMTSImpliedParameters(object):
    documentation = (
        'FX and Equity',
        ['For Risk Neutral simulation, an integrated curve $\\bar{\\sigma}(t)$ needs to be specified and is',
         'interpreted as the average volatility at time $t$. This is typically obtained from the corresponding',
         'ATM volatility. This is then used to construct a new variance curve $V(t)$ which is defined as',
         '$V(0)=0, V(t_i)=\\bar{\\sigma}(t_i)^2 t_i$ and $V(t)=\\bar{\\sigma}(t_n)^2 t$ for $t>t_n$ where',
         '$t_1,...,t_n$ are discrete points on the ATM volatility curve.',
         '',
         'Points on the curve that imply a decrease in variance (i.e. $V(t_i)<V(t_{i-1})$) are adjusted to',
         '$V(t_i)=\\bar\\sigma(t_i)^2t_i=V(t_{i-1})$. This curve is then used to construct *instantaneous* curves',
         'that are then input to the corresponding stochastic process.',
         '',
         'The relationship between integrated $F(t)=\\int_0^t f_1(s)f_2(s)ds$ and instantaneous curves $f_1, f_2$',
         'where the instantaneous curves are defined on discrete points $P={t_0,t_1,..,t_n}$ with $t_0=0$ is defined',
         'on $P$ by Simpson\'s rule:',
         '',
         '$$F(t_i)=F(t_{i-1})+\\frac{t_i-t_{i-1}}{6}\\Big(f(t_i)+4f(\\frac{t_i+t_{i-1}}{2})+f(t_i)\\Big)$$',
         '',
         'and $f(t)=f_1(t)f_2(t)$. Integrated curves are flat extrapolated and linearly interpolated.'
         ]
    )

    def __init__(self, param, prec=np.float32):
        self.prec = prec
        self.param = param

    def bootstrap(self, sys_params, price_models, price_factors, market_prices, calendars, debug=None):
        '''
        Checks for Declining variance in the ATM vols of the relevant price factor and corrects accordingly.
        '''
        for market_price, implied_params in market_prices.items():
            rate = utils.check_rate_name(market_price)
            market_factor = utils.Factor(rate[0], rate[1:])
            if market_factor.type == 'GBMTSModelPrices':
                # get the vol surface
                vol_factor = utils.Factor('FXVol', utils.check_rate_name(
                    implied_params['instrument']['Asset_Price_Volatility']))

                # this shouldn't fail - if it does, need to log it and move on
                try:
                    fxvol = riskfactors.construct_factor(vol_factor, price_factors)
                except Exception:
                    logging.error('Unable to bootstrap {0} - skipping'.format(market_price), exc_info=True)
                    continue

                mn_ix = np.searchsorted(fxvol.moneyness, 1.0)
                atm_vol = [np.interp(1, fxvol.moneyness[mn_ix - 1:mn_ix + 1], y) for y in
                           fxvol.get_vols()[:, mn_ix - 1:mn_ix + 1]]

                # store the output
                price_param = utils.Factor('GBMTSImpliedParameters', market_factor.name)
                model_param = utils.Factor('GBMAssetPriceTSModelImplied', market_factor.name)

                if fxvol.expiry.size > 1:
                    dt = np.diff(np.append(0, fxvol.expiry))
                    var = fxvol.expiry * np.array(atm_vol) ** 2
                    sig = atm_vol[:1]
                    vol = atm_vol[:1]
                    var_tm1 = var[0]
                    fixed_variance = False

                    for var_t, delta_t, t_i in zip(var[1:], dt[1:] / 3.0, fxvol.expiry[1:]):
                        M = var_tm1 + delta_t * (sig[-1] ** 2)
                        if var_t < M:
                            fixed_variance = True
                            var_t = M

                        a = delta_t
                        b = sig[-1] * delta_t
                        c = M - var_t

                        sig.append((-b + np.sqrt(b * b - 4.0 * a * c)) / (2.0 * a))
                        vol.append(np.sqrt(var_t / t_i))
                        var_tm1 = var_t

                    if fixed_variance:
                        logging.warning('Fixed declining variance for {0}'.format(market_price))
                else:
                    vol = atm_vol

                price_factors[utils.check_tuple_name(price_param)] = OrderedDict(
                    [('Property_Aliases', None),
                     ('Vol', utils.Curve(['Integrated'], list(zip(fxvol.expiry, vol)))),
                     ('Quanto_FX_Volatility', None),
                     ('Quanto_FX_Correlation', 0.0)])
                price_models[utils.check_tuple_name(model_param)] = OrderedDict([('Risk_Premium', None)])


class RiskNeutralInterestRateModel(object):
    def __init__(self, param, prec=np.float32):
        self.param = param
        self.num_batches = 1
        self.batch_size = 512 * 10
        self.prec = prec
        self.sample = None
        # set the global precision - not ideal
        utils.Default_Precision = prec

    def calc_loss_on_ir_curve(self, implied_params, base_date, time_grid, process,
                              implied_obj, ir_factor, vol_surface, resid=tf.square, debug=None):
        # calculate a reverse lookup for the tenors and store the daycount code
        all_tenors = utils.update_tenors(base_date, {ir_factor: process})
        # calculate the curve index - need to clean this up - TODO!!!
        curve_index = [instruments.calc_factor_index(ir_factor, {}, {ir_factor: 0}, all_tenors)]
        # calc the market swap rates and instrument_definitions    
        market_swaps, benchmarks = create_market_swaps(
            base_date, time_grid, curve_index, vol_surface, process.factor,
            implied_params['instrument']['Instrument_Definitions'], ir_factor.name)
        # number of random factors to use
        numfactors = process.num_factors()
        # random sample - use a sobol sequence skipping the first 4000 numbers
        # need to check convergence and error - TODO
        # also, this is in python - might want to use cffi to speed this up
        sample = self.calc_sample(time_grid, numfactors)
        tensor_swaptions = {}
        # setup a common context - we leave out the random numbers and pass it in explicitly below
        shared_mem = curve_calibration_class(
            t_Buffer={},
            t_Scenario_Buffer=[None],
            precision=self.prec,
            simulation_batch=self.batch_size)
        # setup the variables
        implied_var = {}
        # the curve is treated as constant here - no placeholders
        stoch_var = tf.constant(process.factor.current_value(), dtype=self.prec)
        with tf.name_scope("Implied_Input"):
            for param_name, param_value in implied_obj.current_value().items():
                factor_name = utils.Factor(
                    implied_obj.__class__.__name__, ir_factor.name + (param_name,))
                tf_variable = tf.get_variable(
                    name=utils.check_scope_name(factor_name),
                    initializer=param_value.astype(self.prec),
                    dtype=self.prec)
                implied_var[param_name] = tf_variable

        # now setup the calc
        process.precalculate(base_date, time_grid, stoch_var, shared_mem, 0, implied_tensor=implied_var)

        # needed to interpolate the zero curve
        delta_scen_t = np.diff(time_grid.scen_time_grid).reshape(-1, 1).astype(self.prec)

        for batch_index in range(self.num_batches):
            # load up the batch
            batch_sample = sample[batch_index].T.reshape(
                numfactors, time_grid.time_grid_years.size, -1).astype(self.prec)
            # simulate the price factor - only need the curve at the mtm time points
            shared_mem.t_Scenario_Buffer[0] = process.generate(shared_mem, batch_sample)
            # get the discount factors
            Dfs = utils.calc_time_grid_curve_rate(
                curve_index, time_grid.calc_time_grid(time_grid.scen_time_grid[:-1]),
                shared_mem, delta_scen_t, multiply_by_time=True)
            # get the index in the deflation factor just prior to the given grid
            deflation = Dfs.reduce_deflate(time_grid.mtm_time_grid, shared_mem)
            # go over the instrument definitions and build the calibration
            for swaption_name, market_data in market_swaps.items():
                expiry = market_data.deal_data.Time_dep.mtm_time_grid[
                    market_data.deal_data.Time_dep.deal_time_grid[0]]
                DtT = deflation[expiry]
                par_swap = pricing.pvfloatcashflowlist(
                    shared_mem, time_grid, market_data.deal_data,
                    pricing.pricer_float_cashflows, settle_cash=False)
                sum_swaption = tf.reduce_sum(tf.nn.relu(DtT * par_swap))
                if swaption_name in tensor_swaptions:
                    tensor_swaptions[swaption_name] += sum_swaption
                else:
                    tensor_swaptions[swaption_name] = sum_swaption

        calibrated_swaptions = {k: v / (self.batch_size * self.num_batches) for k, v in tensor_swaptions.items()}
        error = {k: swap.weight * resid(100.0 * (
                swap.price / calibrated_swaptions[k] - 1.0)) for k, swap in market_swaps.items()}

        return implied_var, error, calibrated_swaptions, market_swaps, benchmarks

    def bootstrap(self, sys_params, price_models, price_factors, market_prices, calendars, debug=None):
        base_date = sys_params['Base_Date']
        base_currency = sys_params['Base_Currency']

        for market_price, implied_params in market_prices.items():
            rate = utils.check_rate_name(market_price)
            market_factor = utils.Factor(rate[0], rate[1:])
            if market_factor.type == self.market_factor_type:
                # fetch the factors
                ir_factor = utils.Factor('InterestRate', rate[1:])
                vol_factor = utils.Factor('InterestYieldVol', utils.check_rate_name(
                    implied_params['instrument']['Swaption_Volatility']))

                # this shouldn't fail - if it does, need to log it and move on
                try:
                    swaptionvol = riskfactors.construct_factor(vol_factor, price_factors)
                    ir_curve = riskfactors.construct_factor(ir_factor, price_factors)
                except KeyError as k:
                    logging.warning('Missing price factor {} - Unable to bootstrap {}'.format(k.args, market_price))
                    continue
                except Exception:
                    logging.error('Unable to bootstrap {0} - skipping'.format(market_price), exc_info=True)
                    continue

                # set of dates for the calibration
                mtm_dates = set(
                    [base_date + x['Start'] for x in implied_params['instrument']['Instrument_Definitions']])

                # grab the implied process
                implied_obj, process, vol_tenors = self.implied_process(
                    base_currency, price_factors, price_models, ir_curve, rate)

                # setup the time grid
                time_grid = TimeGrid(mtm_dates, mtm_dates, mtm_dates)
                # add a delta of 10 days to the time_grid_years (without changing the scenario grid
                # this is needed for stochastically deflating the exposure later on
                time_grid.set_base_date(base_date, delta=(10, vol_tenors*utils.DAYS_IN_YEAR))

                # setup the tensorflow calc
                graph = tf.Graph()
                with graph.as_default():
                    # calculate the error
                    loss, optimizers, implied_var, calibrated_swaptions, market_swaptions, benchmarks = self.calc_loss(
                        implied_params, base_date, time_grid, process, implied_obj, ir_factor, swaptionvol)

                if debug is not None:
                    debug.deals['Deals']['Children'] = [{'instrument': x} for x in benchmarks]
                    try:
                        debug.write_trade_file(market_factor.name[0] + '.aap')
                    except Exception:
                        logging.error('Could not write output file {}'.format(market_factor.name[0] + '.aap'))

                config = tf.ConfigProto(allow_soft_placement=True)

                with tf.Session(graph=graph, config=config) as sess:
                    # init the variables
                    sess.run(tf.global_variables_initializer())
                    # check the time
                    time_now = time.clock()
                    batch_loss, vars = sess.run([loss, implied_var])
                    logging.info('{} - Batch loss {}'.format(market_factor.name[0], batch_loss))
                    for k, v in sorted(vars.items()):
                        logging.info('{} - {}'.format(k, v))

                    sim_swaptions = sess.run(calibrated_swaptions)
                    for k, v in sorted(sim_swaptions.items()):
                        price = market_swaptions[k].price
                        logging.debug('{},market_value,{:f},sim_model_value,{:f},error,{:.0f}%'.format(
                            k, price, v, 100.0 * (price - sim_swaptions[k]) / price))

                    # minimize
                    soln = None
                    num_optimizers = len(optimizers)
                    for op_loop in range(2 * num_optimizers):
                        optimizers[op_loop % num_optimizers].minimize(sess)
                        batch_loss, vars = sess.run([loss, implied_var])

                        if soln is None or batch_loss < soln[0]:
                            soln = (batch_loss, vars)
                            logging.info('{} - run {} - Batch loss {}'.format(
                                market_factor.name[0], op_loop, batch_loss))
                            for k, v in sorted(vars.items()):
                                logging.info('{} - {}'.format(k, v))
                            sim_swaptions = sess.run(calibrated_swaptions)
                            for k, v in sorted(sim_swaptions.items()):
                                price = market_swaptions[k].price
                                logging.info('{},market_value,{:f},sim_model_value,{:f},error,{:.0f}%'.format(
                                    k, price, v, 100.0 * (price - sim_swaptions[k]) / price))

                    # save this
                    self.save_params(soln[1], price_factors, implied_obj, rate)
                    # record the time
                    logging.info('This took {} seconds.'.format(time.clock() - time_now))


class PCAMixedFactorModelParameters(RiskNeutralInterestRateModel):
    def __init__(self, param, prec=np.float32):
        super(PCAMixedFactorModelParameters, self).__init__(param)
        self.market_factor_type = 'HullWhite2FactorInterestRateModelPrices'

    def calc_sample(self, time_grid, numfactors=0):
        if numfactors != 3 or self.sample is None:
            self.sample = normalize(norm.ppf(hdsobol.gen_sobol_vectors(
                self.batch_size * self.num_batches + 4000, time_grid.scen_time_grid.size * numfactors))[3999:]).reshape(
                self.num_batches, self.batch_size, -1)
        return self.sample

    def calc_loss(self, implied_params, base_date, time_grid, process, implied_obj, ir_factor, vol_surface):
        # get the swaption error and market values
        implied_var, error, calibrated_swaptions, market_swaptions, benchmarks = self.calc_loss_on_ir_curve(
            implied_params, base_date, time_grid, process, implied_obj, ir_factor, vol_surface)

        losses = list(error.values())
        loss = tf.reduce_sum(losses)

        loss_with_penatalties = loss + tf.nn.moments(
            implied_var['Yield_Volatility'][1:] - implied_var['Yield_Volatility'][:-1], axes=[0])[1]

        var_to_bounds = {implied_var['Yield_Volatility']: (1e-3, 0.6),
                         implied_var['Reversion_Speed']: (1e-2, 1.8)}

        # optimizer = ScipyLeastsqOptimizerInterface(losses, var_to_bounds={} if var_to_bounds is None else var_to_bounds )
        optimizer = ScipyBasinOptimizerInterface(loss, var_to_bounds=var_to_bounds)

        return loss, optimizer, implied_var, calibrated_swaptions, market_swaptions, benchmarks

    def implied_process(self, base_currency, price_factors, price_models, ir_curve, rate):
        # need to create a process and params as variables to pass to tf
        price_model = utils.Factor('PCAInterestRateModel', rate[1:])
        param_name = utils.check_tuple_name(price_model)
        vol_tenors = np.array([0, 1.5, 3, 6, 12, 24, 48, 84, 120]) / 12.0

        if param_name in price_factors:
            param = price_models[param_name]

            # construct an initial guess - need to read from params
            implied_obj = riskfactors.PCAMixedFactorModelParameters(
                {'Quanto_FX_Volatility': None,
                 'Reversion_Speed': param['Reversion_Speed'],
                 'Yield_Volatility': param['Yield_Volatility']})
        else:
            # need to do a historical calibration
            implied_obj = riskfactors.PCAMixedFactorModelParameters(
                {'Quanto_FX_Volatility': None,
                 'Reversion_Speed': 1.0,
                 'Yield_Volatility': utils.Curve([], list(zip(vol_tenors, [0.1] * vol_tenors.size)))})

        process = stochasticprocess.construct_process(
            price_model.type, ir_curve, price_models[utils.check_tuple_name(price_model)], implied_obj)

        return implied_obj, process, vol_tenors

    def save_params(self, vars, price_factors, implied_obj, rate):
        # construct an initial guess - need to read from params
        param_name = utils.check_tuple_name(
            utils.Factor(type='PCAInterestRateModel', name=rate[1:]))
        param = price_factors[param_name]
        vol_tenors = implied_obj.get_vol_tenors()

        param['Reversion_Speed'] = float(vars['Reversion_Speed'][0])
        param['Yield_Volatility'].array = np.dstack((vol_tenors, vars['Yield_Volatility']))[0]

        # assign this back in case this is a dict-proxy object
        price_factors[param_name] = param


class HullWhite2FactorModelParameters(RiskNeutralInterestRateModel):
    documentation = (
        'Interest Rates',
        ['A set of parameters $\\sigma_1, \\sigma_2, \\alpha_1, \\alpha_2, \\rho$ are estimated from ATM',
         'swaption volatilities. Swaption volatilities are preferred to caplets to better estimate $\\rho$.'
         'Although assuming that $\\sigma_1, \\sigma_2$ are constant makes the calibration of this model',
         'considerably easier, in general, $\\sigma_1, \\sigma_2$ should be allowed a piecewise linear term',
         'structure dependent on the underlying swaptions.',
         '',
         'For a set of $J$ ATM swaptions, we need to minimize:',
         '',
         '$$E=\\sum_{j\\in J} \\omega_j (V_j(\\sigma_1, \\sigma_2, \\alpha_1, \\alpha_2, \\rho)-V_j)^2$$',
         '',
         'Where $V_j(\\sigma_1, \\sigma_2, \\alpha_1, \\alpha_2, \\rho)$ is the price of the $j^{th}$ swaption',
         'under the model, $V_j$ is the market value of the $j^{th}$ swaption and $ \\omega_j$ is the corresponding',
         'weight. The market value is calculated using the standard pricing functions',
         '',
         'To find a good minimum of the model value, basin hopping as implemented [here](https://docs.scipy.org/doc\
/scipy/reference/generated/scipy.optimize.basinhopping.html) as well as',
         'least squares [optimization](https://docs.scipy.org/doc/scipy/reference/generated/\
scipy.optimize.leastsq.html) are used.',
         '',
         'The error $E$ is algorithmically differentiated and then solved via brute-force monte carlo',
         'using tensorflow and scipy.',
         '',
         'If the currency of the interest rate is not the same as the base currency, then a quanto correction needs',
         'to be made. Assume $C$ is the value of the interest rate/FX correlation price factor (can be estimated from',
         'historical data), then the FX rate follows:',
         '',
         '$$d(log X)(t)=(r_0(t)-r(t)-\\frac{1}{2}v(t)^2)dt+v(t)dW(t)$$',
         '',
         'with $r(t)$ the short rate and $r_0(t)$ the short rate in base currency. The short rate with a quanto',
         'correction is:',
         '',
         '$$dr(t)=r_T(0,t)dt+\\sum_{i=1}^2 (\\theta_i(t)-\\alpha_i x_i(t)- \\bar\\rho_i\\sigma_i v(t))dt+\\sigma_i dW_i(t)$$',
         '',
         'where $W_1(t),W_2(t)$ and $W(t)$ are standard Wiener processes under the rate currency\'s risk neutral measure',
         'and $r_T(t,T)$ is the partial derivative of the instantaneous forward rate r(t,T) with respect to the maturity.',
         'date $T$.'
         '',
         'Define:',
         '$$F(u,v)=\\frac{\\sigma_1u+\\sigma_2v}{\\sqrt{\\sigma_1^2+\\sigma_2^2+2\\rho\\sigma_1\\sigma_2}}$$',
         '',
         'Then $\\bar\\rho_1, \\bar\\rho_2$ are assigned:',
         '',
         '$$\\bar\\rho_1=F(1,\\rho)C$$',
         '$$\\bar\\rho_2=F(\\rho,1)C$$',
         '',
         'This is simply assumed to work'
         ]
    )

    def __init__(self, param, prec=np.float32):
        super(HullWhite2FactorModelParameters, self).__init__(param)
        self.market_factor_type = 'HullWhite2FactorInterestRateModelPrices'

    def calc_sample(self, time_grid, numfactors=0):
        if numfactors != 2 or self.sample is None:
            self.sample = normalize(norm.ppf(hdsobol.gen_sobol_vectors(
                self.batch_size * self.num_batches + 4000, time_grid.time_grid_years.size * numfactors))[
                                    3999:]).reshape(
                self.num_batches, self.batch_size, -1)
        return self.sample

    def calc_loss(self, implied_params, base_date, time_grid, process, implied_obj, ir_factor, vol_surface):

        def make_bounds(implied_var, sigma_bounds, corr_bounds, alpha_bounds):
            return {
                implied_var['Sigma_1']: sigma_bounds,
                implied_var['Sigma_2']: sigma_bounds,
                implied_var['Correlation']: corr_bounds,
                implied_var['Alpha_1']: alpha_bounds,
                implied_var['Alpha_2']: alpha_bounds}

        def split_param(x):
            corr = x[-1:]
            alpha = x[-3:-1]
            sigmas = x[:-3]
            return sigmas, alpha, corr

        def make_basin_callbacks(step, sigma_min_max, alpha_min_max, corr_min_max):
            def bounds_check(**kwargs):
                x = kwargs["x_new"]
                sigmas, alpha, corr = split_param(x)
                sigma_ok = (sigmas > sigma_min_max[0]).all() and (sigmas < sigma_min_max[1]).all()
                alpha_ok = (alpha > alpha_min_max[0]).all() and (alpha < alpha_min_max[1]).all()
                corre_ok = (corr > corr_min_max[0]).all() and (corr < corr_min_max[1]).all()
                return sigma_ok and alpha_ok and corre_ok

            def basin_step(x):
                sigmas, alpha, corr = split_param(x)
                # update vars
                sigmas = (sigmas * np.exp(np.random.uniform(-step, step, sigmas.size))).clip(*sigma_min_max)
                alpha = (alpha * np.exp(np.random.uniform(-step, step, alpha.size))).clip(*alpha_min_max)
                corr = (corr + np.random.uniform(-step, step, corr.size)).clip(*corr_min_max)

                return np.concatenate((sigmas, alpha, corr))

            return bounds_check, basin_step

        # get the swaption error and market values
        implied_var_dict, error_dict, calibrated_swaptions, market_swaptions, benchmarks = self.calc_loss_on_ir_curve(
            implied_params, base_date, time_grid, process, implied_obj, ir_factor, vol_surface)

        error = OrderedDict(error_dict)
        var_list = [implied_var_dict['Sigma_1'], implied_var_dict['Sigma_2'],
                    implied_var_dict['Alpha_1'], implied_var_dict['Alpha_2'],
                    implied_var_dict['Correlation']]

        implied_var = OrderedDict(implied_var_dict)
        losses = list(error.values())
        loss = tf.reduce_sum(losses)

        sigma_bounds = (1e-4, 0.09)
        alpha_bounds = (1e-4, 2.4)
        corr_bounds = (-.92, 0.92)

        # slightly different bounds for the least squares calc
        lq_sigma_bounds = (1e-5, 0.095)
        lq_alpha_bounds = (1e-5, 2.5)
        lq_corr_bounds = (-.93, 0.93)

        var_to_bounds = make_bounds(implied_var, sigma_bounds, corr_bounds, alpha_bounds)
        var_to_bounds_lq = make_bounds(implied_var, lq_sigma_bounds, lq_corr_bounds, lq_alpha_bounds)

        bounds_ok, make_step = make_basin_callbacks(0.25, sigma_bounds, alpha_bounds, corr_bounds)

        optimizers = [
            ScipyBasinOptimizerInterface(
                loss, var_list=var_list,
                take_step=make_step,
                accept_test=bounds_ok,
                var_to_bounds=var_to_bounds)
            , ScipyLeastsqOptimizerInterface(
                losses, var_list=var_list,
                var_to_bounds=var_to_bounds_lq)
        ]

        return loss, optimizers, implied_var, calibrated_swaptions, market_swaptions, benchmarks

    def implied_process(self, base_currency, price_factors, price_models, ir_curve, rate):
        vol_tenors = np.array([0, 1, 2, 3, 6, 12, 24, 48, 84, 120]) / 12.0
        # construct an initial guess - need to read from params
        param_name = utils.check_tuple_name(
            utils.Factor(type='HullWhite2FactorModelParameters', name=rate[1:]))

        # check if we need a quanto fx vol
        fx_factor = utils.Factor('GBMTSImpliedParameters', ir_curve.get_currency())
        ir_factor = utils.Factor('InterestRate', ir_curve.get_currency())
        fx_factor_name = utils.check_tuple_name(fx_factor)
        ir_factor_name = utils.check_tuple_name(ir_factor)

        if fx_factor_name in price_factors:
            quanto_fx = price_factors[fx_factor_name]['Vol']
            curr_pair = sorted((base_currency,) + ir_curve.get_currency())
            correlation_name = 'Correlation.FxRate.{}/{}'.format('.'.join(curr_pair), ir_factor_name)
            # check if the quote is against the base currency
            sign = 1.0
            if curr_pair[0] == base_currency:
                sign = -1.0
                logging.info('Reversing Correlation as {} is quoted against the base currency'.format(correlation_name))
            # the correlation between fx and ir - needed to establish Quanto Correlation 1 and 2
            C = sign * price_factors.get(correlation_name, {'Value': 0.0})['Value']
        else:
            C = None
            quanto_fx = None

        if param_name in price_factors:
            param = price_factors[param_name]
            implied_obj = riskfactors.HullWhite2FactorModelParameters(
                {'Quanto_FX_Volatility': quanto_fx,
                 'short_rate_fx_correlation': C,
                 'Alpha_1': np.clip(param['Alpha_1'], 1e-4, 2.4),
                 'Alpha_2': np.clip(param['Alpha_2'], 1e-4, 2.4),
                 'Correlation': np.clip(param['Correlation'], -0.95, 0.95),
                 'Sigma_1': utils.Curve([], list(zip(
                     vol_tenors, np.interp(vol_tenors, *param['Sigma_1'].array.T).clip(1e-4 + 5e-5, 0.09)))),
                 'Sigma_2': utils.Curve([], list(zip(
                     vol_tenors, np.interp(vol_tenors, *param['Sigma_2'].array.T).clip(1e-4 + 5e-5, 0.09))))})
        else:
            implied_obj = riskfactors.HullWhite2FactorModelParameters(
                {'Quanto_FX_Volatility': quanto_fx,
                 'short_rate_fx_correlation': C,
                 'Alpha_1': 0.1, 'Alpha_2': 0.1, 'Correlation': 0.01,
                 'Sigma_1': utils.Curve([], list(zip(vol_tenors, [0.01] * vol_tenors.size))),
                 'Sigma_2': utils.Curve([], list(zip(vol_tenors, [0.01] * vol_tenors.size)))})

        # need to create a process and params as variables to pass to tf
        process = stochasticprocess.HullWhite2FactorImpliedInterestRateModel(
            ir_curve, {'Lambda_1': 0.0, 'Lambda_2': 0.0}, implied_obj)

        return implied_obj, process, vol_tenors

    def save_params(self, vars, price_factors, implied_obj, rate):
        param_name = utils.check_tuple_name(
            utils.Factor(type='HullWhite2FactorModelParameters', name=rate[1:]))
        # grab the sigma tenors
        sig1_tenor, sig2_tenor = implied_obj.get_vol_tenors()
        # grab the quanto fx correlations
        quanto_fx1, quanto_fx2 = implied_obj.get_quanto_correlation(
            vars['Correlation'], [vars['Sigma_1'], vars['Sigma_2']])

        param = {'Property_Aliases': None,
                 'Quanto_FX_Volatility': implied_obj.param['Quanto_FX_Volatility'],
                 'Alpha_1': float(vars['Alpha_1'][0]),
                 'Sigma_1': utils.Curve([], list(zip(sig1_tenor, vars['Sigma_1']))),
                 'Quanto_FX_Correlation_1': quanto_fx1,
                 'Alpha_2': float(vars['Alpha_2'][0]),
                 'Sigma_2': utils.Curve([], list(zip(sig2_tenor, vars['Sigma_2']))),
                 'Quanto_FX_Correlation_2': quanto_fx2,
                 'Correlation': float(vars['Correlation'][0])}

        price_factors[param_name] = param


def construct_bootstrapper(btype, param):
    return globals().get(btype)(param)



import os
import utils
import logging
import numpy as np
import pandas as pd
import collections.abc


def update_dict(d, u):
    for k, v in u.items():
        d[k] = update_dict(d.get(k, {}), v) if isinstance(v, collections.abc.Mapping) else v
    return d


def diag_ir(out, calc, factor, tenor_point, aa=None):
    ir = out['Results']['scenarios'][factor]
    size = ir.shape[0]
    perc = np.array([50, 0.50, 2.50, 5.00, 95.00, 97.50, 99.50])
    ten_ir = calc.all_factors[factor].factor.tenors
    all = [pd.DataFrame(np.percentile(i, 100.0 - perc, axis=1), index=perc, columns=ten_ir).T for i in ir]
    comp = pd.DataFrame([x.iloc[ten_ir.searchsorted(tenor_point)] for x in all],
                        index=calc.time_grid.time_grid_years[:size])
    if aa is not None:
        return comp.reindex(comp.index.union(aa['Time (Years)'])).interpolate('index').reindex(aa['Time (Years)'])
    else:
        return comp


def makeflatcurve(curr, bps, tenor=30):
    return {'Currency': curr, 'Curve': utils.Curve([], [[0, bps * 0.01 * 0.01], [tenor, bps * 0.01 * 0.01]]),
            'Day_Count': 'ACT_365', 'Property_Aliases': None, 'Sub_Type': 'None'}


def getpath(pathlist, uat=False):
    for path in pathlist:
        if os.path.isdir(path):
            return os.path.join(path, 'UAT') if uat else path


def set_collateral(cx, Agreement_Currency, Opening_Balance, Received_Threshold=0.0, Posted_Threshold=0.0,
                   Minimum_Received=100000.0, Minimum_Posted=100000.0, Liquidation_Period=10.0):
    cx.deals['Deals']['Children'][0]['instrument'].field.update(
        {'Agreement_Currency': Agreement_Currency, 'Opening_Balance': Opening_Balance,
         'Apply_Closeout_When_Uncollateralized': 'No', 'Collateralized': 'True', 'Settlement_Period': 0.0,
         'Balance_Currency': 'ZAR', 'Liquidation_Period': Liquidation_Period,
         'Credit_Support_Amounts':
             {'Received_Threshold': utils.CreditSupportList({1: Received_Threshold}),
              'Minimum_Received': utils.CreditSupportList({1: Minimum_Received}),
              'Posted_Threshold': utils.CreditSupportList({1: Posted_Threshold}),
              'Minimum_Posted': utils.CreditSupportList({1: Minimum_Posted})
              }
         }
    )


def load_market_data(rundate, path, json_name='MarketData.json', setup_funding=False, cva_default=True):
    from adaptiv import AdaptivContext

    context = AdaptivContext()
    context.parse_json(os.path.join(path, rundate, json_name))
    context.parse_calendar_file(os.path.join(path, 'calendars.cal'))

    context.params['System Parameters']['Base_Date'] = pd.Timestamp(rundate)

    if cva_default:
        context.params['Price Factors']['SurvivalProb.DEFAULT'] = {
            'Recovery_Rate': 0.5,
            'Curve': utils.Curve(
                [], [[0.0, 0.0], [.5, .01], [1, .02], [3, .07], [5, .15], [10, .35], [20, .71], [30, 1.0]]),
            'Property_Aliases': None}

    if setup_funding:
        context.params['Price Factors']['InterestRate.ZAR-SWAP.OIS'] = makeflatcurve('ZAR', -15)
        context.params['Price Factors']['InterestRate.ZAR-SWAP.FUNDING'] = makeflatcurve('ZAR', 10)
        context.params['Price Factors']['InterestRate.USD-LIBOR-3M.FUNDING'] = makeflatcurve('USD', 65)

    return context


def run_baseval(context, overrides=None):
    from calculation import construct_calculation
    calc_params = context.deals['Calculation']

    rundate = calc_params['Base_Date'].strftime('%Y-%m-%d')
    params_bv = {'calc_name': ('test1',), 'Run_Date': rundate,
                 'Currency': calc_params['Currency'], 'Greeks': 'No'}

    if overrides is not None:
        update_dict(params_bv, overrides)

    calc = construct_calculation('Base_Revaluation', cx, prec=np.float64)
    out = calc.execute(params_bv)
    return calc, out


def run_cmc(context, overrides=None, prec=np.float32, CVA=True, FVA=False, CollVA=False):
    from calculation import construct_calculation
    calc_params = context.deals['Calculation']

    rundate = calc_params['Base_Date'].strftime('%Y-%m-%d')
    time_grid = str(calc_params['Base_Time_Grid'])

    default_cva = {'Deflate_Stochastically': 'Yes', 'Stochastic_Hazard_Rates': 'No', 'Counterparty': 'DEFAULT'}
    cva_sect = context.deals.get('Calculation', {'Credit_Valuation_Adjustment': default_cva}).get(
        'Credit_Valuation_Adjustment', default_cva)

    params_mc = {'calc_name': ('test1',), 'Time_grid': time_grid, 'Run_Date': rundate,
                 'Currency': calc_params['Currency'], 'Simulation_Batches': 10, 'Batch_Size': 64 * 8,
                 'Random_Seed': 8312, 'Calc_Scenarios': 'No', 'Generate_Cashflows': 'No', 'Partition': 'None',
                 'Generate_Slideshow': 'No', 'PFE_Recon_File': '', 'Dynamic_Scenario_Dates': 'No',
                 'Deflation_Interest_Rate': calc_params['Deflation_Interest_Rate'],
                 # 'Debug': 'G:\\Credit Quants\\CRSTAL\\riskflow\\logs', 'NoModel': 'Constant',
                 'Debug': 'No', 'Tenor_Offset': 0.0,
                 # 'NoModel':'RiskNeutral',
                 'CVA': {'Deflate_Stochastically': cva_sect['Deflate_Stochastically'],
                         'Stochastic_Hazard': cva_sect['Stochastic_Hazard_Rates'],
                         'Counterparty': cva_sect['Counterparty'],
                         # brave choices these . . .
                         'Gradient': 'No', 'Hessian': 'No'},
                 'FVA': {'Funding_Interest_Curve': 'USD-LIBOR-3M.FUNDING',
                         'Risk_Free_Curve': 'USD-OIS',
                         'Counterparty': cva_sect['Counterparty'],
                         'Stochastic_Funding': 'Yes'},
                 'CollVA': {'Gradient': 'Yes'}
                 }

    if overrides is not None:
        update_dict(params_mc, overrides)

    if not CVA:
        del params_mc['CVA']
    if not FVA:
        del params_mc['FVA']
    if not CollVA:
        del params_mc['CollVA']

    calc = construct_calculation('Credit_Monte_Carlo', context, prec=prec)
    out = calc.execute(params_mc)
    exposure = out['Results']['mtm'].clip(0.0, np.inf)
    dates = np.array(sorted(calc.time_grid.mtm_dates))[
        calc.netting_sets.sub_structures[0].obj.Time_dep.deal_time_grid]

    res = pd.DataFrame({'EE': np.mean(exposure, axis=1), 'PFE': np.percentile(exposure, 95, axis=1)}, index=dates)

    # useful to compare against adaptiv
    # comp = diag_ir(out, calc, Factor(type='InterestRate', name=('ZAR-SWAP',)), 2, aa=None)
    return calc, out, res


def bootstrap(path, rundate, reuse_cal=True):
    from adaptiv import AdaptivContext

    context = AdaptivContext()

    if reuse_cal and os.path.isfile(os.path.join(path, rundate, 'CVAMarketData.json')):
        context.parse_json(os.path.join(path, rundate, 'CVAMarketData.json'))
    else:
        context.parse_json(os.path.join(path, rundate, 'MarketData.json'))

    context.params['System Parameters']['Base_Date'] = pd.Timestamp(rundate)
    context.parse_calendar_file(os.path.join(path, 'calendars.cal'))

    logging.basicConfig(level=logging.INFO,
                        format='%(asctime)s %(levelname)-8s %(message)s',
                        datefmt='%m-%d %H:%M')

    context.bootstrap()
    context.write_marketdata_json(os.path.join(path, rundate, 'MarketDataCal.json'))
    context.write_market_file(os.path.join(path, rundate, 'MarketDataCal.dat'))


if __name__ == '__main__':
    import matplotlib

    # matplotlib.use('Qt4Agg')
    import matplotlib.pyplot as plt

    plt.interactive(True)
    # make pandas pretty print
    pd.options.display.float_format = '{:,.5f}'.format

    folder = 'CVA'
    path = getpath(['E:\\Data\\crstal\\{}'.format(folder),
                    'G:\\Credit Quants\\CRSTAL\\{}'.format(folder),
                    'G:\\{}'.format(folder)], uat=False)

    rundate = '2020-03-06'

    # set the visible GPU
    os.environ['CUDA_VISIBLE_DEVICES'] = '-1'
    # set the log level
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

    # bootstrap(path, rundate, reuse_cal=True)

    if 1:
        cx = load_market_data(rundate, path, json_name='MarketData.json')
        cx_new = load_market_data(rundate, path, json_name='CVAMarketData_Calibrated_New.json')

        for factor in [x for x in cx_new.params['Price Factors'].keys()
                       if x.startswith('HullWhite2FactorModelParameters')]:
            # override it
            cx.params['Price Factors'][factor] = cx_new.params['Price Factors'][factor]

        cx.parse_json(os.path.join(path, rundate, 'CrB_Kathu_Solar_Park_ISDA.json'))

        if 1:
            calc, out, res = run_cmc(cx, overrides={'Calc_Scenarios': 'No',
                                                    'Dynamic_Scenario_Dates': 'No',
                                                    #'Run_Date': '2020-03-09',
                                                    #'Tenor_Offset': -3/365.0,
                                                    # 'Time_grid':'1m 5m 1362d',
                                                    'Batch_Size': 64 * 4,
                                                    'Simulation_Batches': 20,
                                                    'CVA': {'Gradient': 'Yes', 'Hessian': 'Yes'}}, prec=np.float32)

            # calc, out, res = run_cmc(cx, rundate)
        else:
            calc, out = run_baseval(cx)
